#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# unittestClipboard.py
#
# this module tests the clipboard module, natlinkclipboard in natlink/macrosystem/core,
# as developed by Christo Butcher clipboard.py for Dragonfly and
# enhanced by Quintijn Hoogenboom, 2019.

#
import sys
import unittest
import types
import os
import os.path
import time
import win32gui
from pathlib import Path

thisDir = Path('.')

import TestCaseWithHelpers
import natlink
from dtactions import natlinkclipboard
from dtactions.unimacro import unimacroactions as actions
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro import unimacroutils
from natlinkcore import natlinkutils

class TestError(Exception):
    pass

natconnectOption = 0 # no threading has most chances to pass...




logFileName = thisDir/"testresult.txt"
print('printing will go to %s'% logFileName)
print('start unittestClipboard', file=open(logFileName, 'w'))

testFilesDir = thisDir/'test_clipboardfiles'
if not testFilesDir.is_dir():
    testFilesDir.mkdir()

#---------------------------------------------------------------------------
# These tests should be run after we call natConnect
# no reopen user at each test anymore..
# no default open window (open window will be the calling program)
# default .ini files pop up when you first run this test. just ignore them.
# the recording of print presents problems.
# All should go to testresult.txt in this same directory
class UnittestClipboard(TestCaseWithHelpers.TestCaseWithHelpers):
    def setUp(self):
        self.connect()  # switched off

        self.thisHndle = win32gui.GetForegroundWindow()
        self.org_text = "xyz"*3
        natlinkclipboard.Clipboard.set_system_text(self.org_text)
        # self.setupWindows()
        self.tempFileHndles = []
        self.killActions = {}   # put a teardown action here if {alt+f4} is not OK
        self.modInfos = {}      # the module info of each hndle (tuple of length 3)
        # possible handles in different tests...
        self.docx0Hndle = self.docx1Hndle = self.docx2Hndle = None
        self.text0Hndle = self.text1Hndle = self.text2Hndle = None
        self.expl0Hndle = None
        self.thundHndle = None   # thunderbird new mail message
        self.frescHndle = None   # Frescobaldi (music type setting)
        
    def tearDown(self):
        for hndle in self.tempFileHndles:
            if not win32gui.IsWindow(hndle):
                print("???window does not exist: %s"% hndle)
                continue
            if hndle == self.thisHndle:
                # print('window hndle %s may not match "thisHndle": %s'% (hndle, self.thisHndle))
                continue
            # print('close window with hndle: %s'% hndle)
            unimacroutils.SetForegroundWindow(hndle)
            curHndle = unimacroutils.GetForegroundWindow()

            if hndle == curHndle:
                if hndle in self.killActions:
                    # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
                    # place to break in debug mode
                    # natlinkutilsqh.SetForegroundWindow(curHndle)
                    action(self.killActions[hndle], modInfo=self.modInfos[hndle])
                else:
                    sendkeys("{alt+f4}")
        natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        self.disconnect()  # disabled, natConnect
        notClosedHndles = []
        for hndle in self.tempFileHndles:
            if hndle and win32gui.IsWindow(hndle):
                notClosedHndles.append(hndle)
    
        if self.frescHndle:
            if self.frescHndle in notClosedHndles:
                notClosedHndles.remove(self.frescHndle)
            else:
                raise TestError('Fresobaldi should not be closed after the test')
        if notClosedHndles:
            raise TestError('Hndles not close, but should: %s'% notClosedHndles)

    def connect(self):
        # start with 1 for thread safety when run from pythonwin:
        # natlink.natConnect(natconnectOption)
        # sys.stdout = open(logFileName, 'a')
        # sys.stderr = open(logFileName, 'a')
        self.isConnected = False
        pass
    
    def disconnect(self):
        # natlink.natDisconnect()
        self.isConnected = False
        pass

    def getTextViaClipboard(self, waitTime):
        """get text of window and selection or cursor via copy paste actions
        
        pass waitTime for testing response time.
        
        waiting_iterations is default 10, but changed to 3 if an empty clipboard is expected
        
        return: text, selStart, selEnd, cursorPos
               
        """
        cb = natlinkclipboard.Clipboard(save_clear=True, debug=1)
        waitTime = 0.001
      
        # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        # natlinkutilsqh.SetForegroundWindow(self.testHndle)
        
        ## first try if there is a selection:
        action("{ctrl+c}")
        selection = cb.get_text(waitTime, waiting_iterations=3)
        if selection:
            print('selection: %s'% selection)
            action("{left}")
            cb.clear_clipboard()

           
        # now try text left of selection:
        action("{ctrl+shift+home}{ctrl+c}")
     
        # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        # natlinkutilsqh.SetForegroundWindow(self.testHndle)
     
     
        before = cb.get_text(waitTime)
        if before:
            action("{right}")
            cb.clear_clipboard()
        else:
            print("at start of document")

        # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        # natlinkutilsqh.SetForegroundWindow(self.testHndle)

        text = before + selection
        selStart = len(before)
        selEnd = len(text)
    
        # skip to end of selection:       
        if selection:
            action("{right %s}"% len(selection))

        # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        # natlinkutilsqh.SetForegroundWindow(self.testHndle)
        
        if text:
            # now select to end:
            action("{shift+ctrl+end}{ctrl+c}")
            after = cb.get_text(waitTime)
            print("after: %s"% repr(after))
            
            if not after:
                if self.testHndle == self.thundHndle:
                    ## after an empty selection, go one to the left!!
                    action("{left}")
                    pass
                after = ""
            else:
                # undo selection
                action("{left}")
            if after.endswith('\0'):
                after = after.replace('\0', '')
                print("removed null char at end of selection (1): %s"% after)
                # after = afterplusone[1:]
        else:
            # presumably at start of buffer
            action("{shift+ctrl+end}{ctrl+c}")
            after = cb.get_text(waitTime, waiting_iterations=3)
            if after:
                # undo selection 
                action("{left}")
            else:
                pass
            if after.endswith('\0'):
                after = after.replace('\0', '')
                print("removed null char at end of selection (2): %s"% after)
                # after = afterplusone[1:]
                cb.clear_clipboard()

        if selection:
            lensel = len(selection)
            # select from left to right;
            
            # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
            # natlinkutilsqh.SetForegroundWindow(self.testHndle)
            
            action("{left %s}{shift+right %s}"% (lensel, lensel))

        text = text + after
        cursorPos = selEnd            

        # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        # natlinkutilsqh.SetForegroundWindow(self.testHndle)


        return text, selStart, selEnd, cursorPos
        

        

    def setupWindows(self):
        """make several windows to which can be switched, and which can be copied from
        
        TODOQH: make bringup of a folder and copy paste of the folder info work... (if somebody wants this)
        """
        return
        self.thisHndle = natlinkutilsqh.GetForegroundWindow()

        dirWindows = "C:\\windows"
        result = actions.AutoHotkeyBringUp(app="explore", filepath=dirWindows)
        if not result:
            print('no result for %s'% dirWindows)
            return
        pPath, wTitle, hndle = result
        self.tempFileHndles.append(hndle)
        self.expl0Hndle = hndle
        natlinkutilsqh.SetForegroundWindow(self.thisHndle)

    def setupDocxFile(self):
        """open Word file and do the testing on it...
        """
        docxFile2 = "natlink.docx"
        docxPath2 = os.path.normPath( os.path.join(testFilesDir, docxFile2))
        if not os.path.isfile(docxPath2):
            raise OSError('file does not exist: %s'% docxPath2)
        result = actions.AutoHotkeyBringUp(app=None, filepath=docxPath2)
        if result:
            pPath, wTitle, hndle = result
        else:
            raise TestError("AutoHotkeyBringUp of Word gives no result: %s"% result)
        self.tempFileHndles.append(hndle)
        natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        print('testempty.docx (docx2Hndle): %s'% result)
        self.tempFileHndles.append(result)
        self.docx2Hndle = result        
        
    def setupTextFiles(self):
        """make some text files for clipboard testing
        """
        textFile0 = "testempty.txt"
        textPath0 = os.path.join(testFilesDir, textFile0)
        open(textPath0, 'w')
        result = actions.AutoHotkeyBringUp(app=None, filepath=textPath0)
        pPath, wTitle, hndle = result
        natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        print('testempty (text0Hndle): %s'% hndle)
        self.tempFileHndles.append(hndle)
        self.text0Hndle = hndle

        textFile1 = "testsmall.txt"
        textPath1 = os.path.join(testFilesDir, textFile1)
        self.text1Txt = "small abacadabra\n"*2
        open(textPath1, 'w').write(self.text1Txt)
        result = actions.AutoHotkeyBringUp(app=None, filepath=textPath1)
        pPath, wTitle, hndle = result
        self.tempFileHndles.append(hndle)
        natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        print('testsmall (text1Hndle): %s'% hndle)
        self.text1Hndle = hndle
        self.tempFileHndles.append(hndle)

        textFile2 = "testlarge.txt"
        textPath2 = os.path.join(testFilesDir, textFile2)
        self.text2Txt = "large abacadabra\n"*1000
        open(textPath2, 'w').write(self.text2Txt)
        result = actions.AutoHotkeyBringUp(app=None, filepath=textPath2)
        pPath, wTitle, hndle = result
        self.tempFileHndles.append(hndle)
        natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        print('testlarge (text2Hndle): %s'% hndle)
        self.text2Hndle = hndle
        self.tempFileHndles.append(hndle)

    def setupThunderbirdNewWindow(self):
        """start an empty mail message in Thunderbird
        
        at shutdown do not kill this window 
        """
        ## this extraaow, and goes to the message pane:
        extra = '\n'.join(['Send, ^n',
                'SetTitleMatchMode, 2',
                'WinWaitNotActive, ##title##',
                'Sleep, 100',
                'Send, {tab 2}'])     # note SetTitleMatchMode and Sleep, apparently necessary

        result = actions.AutoHotkeyBringUp(app="thunderbird.exe", title="Mozilla Thunderbird", extra=extra)
        pPath, wTitle, hndle = result
        self.tempFileHndles.append(hndle)
        self.thundHndle = hndle
        self.killActions[hndle] = "KW({ctrl+w}, w)"   # kill the window, with letter w if windows has text....
        self.modInfos[hndle] = result
        action("Thunderbird new window abacadabra")

    def setupFrescobaldiNewPane(self):
        """start an empty pane in Frescobaldi (front end for Lilypond music type setter)
        
        at shutdown do not kill this window
        """
        ## this extra makes an empty window, and goes to the message pane:
        extra = '\n'.join(["Sleep, 100", 'Send, ^n', "Sleep, 100"])

        result = actions.AutoHotkeyBringUp(app=r"C:\Program Files (x86)\Frescobaldi\frescobaldi.exe", title="Frescobaldi", extra=extra)
        pPath, wTitle, hndle = result
        action("Frescobaldi abacadabra")
        self.tempFileHndles.append(hndle)
        self.frescHndle = hndle
        self.killActions[hndle] = "KW({ctrl+w}, {right}{enter})"
        self.modInfos[hndle] = result

    def testEmpty(self):
        """only testing the setup and teardown
        
        implicitly testing actions (via autohotkey, AHK) like starting a process and switching to a window handle
        
        (this presented a lot of trouble, getting the correct files open with the correct window handles)
        
        Now with AHK on, this runs more smooth than ever. With or without Dragon running.
        """
        print("doing an empty test, just to test setUp and tearDown")
        self.setupTextFiles()
        # self.setupDocxFile()
        # self.setupWindows()
        # self.setupThunderbirdNewWindow()
        # self.setupFrescobaldiNewPane()
        natlinkutilsqh.SetForegroundWindow(self.thisHndle)
        # print("Frescobaldi handle: ", self.frescHndle)
        # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
                
    def tttestCopyClipboardGetSetText(self):
        """use the clipboard for getting and the window text
        
        Run with one target at a time
        
        
        """
        print("testCopyClipboardSwitching")
        # self.setupDocxFile()
        self.setupThunderbirdNewWindow()
        self.testHndle = self.thundHndle
        # self.setupFrescobaldiNewPane()
        # self.testHndle = self.frescHndle
        # self.setupTextFiles()
        
        
        for i in range(1, 21, 10):
            waitTime = 0.001/i

            t0 = time.time()
    
            ## start with empty window
            action("{ctrl+a}{del}")
            
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime=waitTime)
            expText = ""
            expPosTuple = (0,0,0)  # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")
            
            action("Hello world")

            # these pair of lines provide the possibility to debug the tests
            # put the breakpoint at the second line...
            # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
            # natlinkutilsqh.SetForegroundWindow(self.testHndle)
    
    
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            expText = "Hello world"
            expPosTuple = (11, 11, 11)  # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")
            pass

    
            # now select word world
            action("{shift+ctrl+left}")
    
            # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
            # natlinkutilsqh.SetForegroundWindow(self.testHndle)

    
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            expText = "Hello world"
            expPosTuple = (6, 11, 11)  # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")

            ## and again:
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")
    
            # natlinkutilsqh.SetForegroundWindow(self.thisHndle)
            # natlinkutilsqh.SetForegroundWindow(self.testHndle)
    
            # now replace the word world, and make a second line
            action('SCLIP(WORLD{enter}How are you going?{enter}In this Corona world?)')
            expText = "Hello WORLD\nHow are you going?\nIn this Corona world?"
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            expPosTuple = (52, 52, 52)  # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")
            pass

            ## and again:    
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")
    
            # select all
            action("{ctrl+home}{shift+ctrl+end}")
            expText = "Hello WORLD\nHow are you going?\nIn this Corona world?"
            # if self.testHndle == self.thundHndle:
            #     expText += "\n"
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            expPosTuple = (0, len(expText), len(expText))  # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")

            # and again:
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            expPosTuple = (0, len(expText), len(expText))  # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")

            # select "are you"
            action("{ctrl+home}{down}{right 4}{shift+right 7}")
            expText = "Hello WORLD\nHow are you going?\nIn this Corona world?"
            # not here!!
            # if self.testHndle == self.thundHndle:
            #     expText += "\n"
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            expPosTuple = (16, 23, 23) # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")
    
            ## and again:
            text, startSel, endSel, cursorPos = self.getTextViaClipboard(waitTime)
            expPosTuple = (16, 23, 23) # startSel, endSel, cursorPos
            self.assert_equal(expText, text, "text not as expected")
            self.assert_equal(expPosTuple, (startSel, endSel, cursorPos), "positions not as expected")


            t1 = time.time()
            elapsed = t1 - t0
            print("time %.6f for round with waitTime: %.6f "% (elapsed, waitTime))


    def tttestCopyClipboardSwitching(self):
        """test the copying of the clipboard
        
        with empty, small and larger .txt file.
        
        .docx does not run smooth
        
        
        """
        print("testCopyClipboardSwitching")
        # self.setupDocxFile()
        # self.setupWindows()
        # self.setupThunderbirdNewWindow()
        self.setupFrescobaldiNewPane()
        self.setupTextFiles()

        cb = natlinkclipboard.Clipboard(save_clear=True, debug=2)   # debug flag can be set or omitted (or set to 0)
        return

        ## longer test:
        for i in range(1, 10, 5):
            waitTime = 0.001/i
            print('---- round: %s, waitTime %.4f '% (i, waitTime))
            
            ## first clear the clipboard:
            cb.clear_clipboard()
            got = cb.get_text(waitTime)
            print('after clear_clipboard, text: %s (cb: %s)'% (got, cb))
            self.assert_equal("", got, "should have no text now, waitTime: %s (times 4)"% waitTime)

            print("now testempty.txt -----------------")
            # empty file: 
            natlinkutilsqh.SetForegroundWindow(self.text0Hndle)
            sendkeys("{ctrl+a}{ctrl+c}")
            got = cb.get_text(waitTime)
            print('after select all copy of empty file, no change should have happened:')
            # print('got: %s, (cb: %s)'% (got, cb))
            self.assert_equal("", got, "should have no text now, waitTime: %s"% waitTime)
            
            natlinkutilsqh.SetForegroundWindow(self.thisHndle)

            print("now testsmall.txt -----------------")
            natlinkutilsqh.SetForegroundWindow(self.text1Hndle)
            sendkeys("{ctrl+a}{ctrl+c}")
            got = cb.get_text(waitTime)
            print('after select all copy of testsmall.txt')
            # print('got: %s, (cb: %s)'% (got, cb))
            exp = self.text1Txt
            self.assert_equal(exp, got, "testsmall should have two lines of text, waitTime: %s"% waitTime)
            
            
            # test large.txt
            print("now testlarge.txt -----------------")
            natlinkutilsqh.SetForegroundWindow(self.text2Hndle)
            sendkeys("{ctrl+a}{ctrl+c}")
            cb.get_text(waiting_interval=waitTime)
            got = cb.get_text()
            if got:
                lengot = len(got)
            else:
                print('testlarge, no result from get_text: %s'% cb)
            exp = len(self.text2Txt)
            self.assert_equal(exp, lengot, "should have long text now, waitTime: %s"% waitTime)
            # empty for the next round:
            # cb.set_text("")
            time.sleep(waitTime)

            natlinkutilsqh.SetForegroundWindow(self.thisHndle)

            # test natlink.docx
            print("now natlink.docx (%s)-----------------"% self.docx2Hndle)
            if not self.docx2Hndle:
                print("word document not available: %s"% (self.docx2Hndle))
                continue
            natlinkutilsqh.SetForegroundWindow(self.docx2Hndle)
            sendkeys("{ctrl+a}{ctrl+c}")
            cb.get_text(waiting_interval=waitTime)
            got = cb.get_text()
            if got:
                lengot = len(got)
            else:
                print('testlarge, no result from get_text: %s'% cb)
            exp = len(self.text2Txt)
            self.assert_equal(exp, lengot, "should have long text now, waitTime: %s"% waitTime)
            # empty for the next round:
            # cb.set_text("")
            time.sleep(waitTime)
            
        del cb
        got_org_text = natlinkclipboard.Clipboard.get_system_text()
        self.assert_equal(self.org_text, got_org_text, "restored text from clipboard not as expected")

    
    def tttestSetForegroundWindow(self):
        """test switching the different windows, including this
        
        This is a side test for getting the testing system at work...
        
        and should work without problems with the AutoHotkey script...
        
        About 0.15 seconds to get a window in the foreground...
        """
        # self.setupDocxFile()
        self.setupTextFiles()            # self.text0Hndle, ..., self.text2Hndle
        self.setupWindows()
        self.setupFrescobaldiNewPane()   # self.frescHndle
        self.setupThunderbirdNewWindow() # self.thundHndle
        
        thisHndle = natlinkutilsqh.GetForegroundWindow()
        print("thisHndle: %s"% thisHndle)
        unknownHndle = 5678659
        hndles = [thisHndle, self.docx2Hndle, self.docx1Hndle, self.text0Hndle, self.text1Hndle, self.text2Hndle,
                  self.frescHndle, self.thundHndle, unknownHndle]
        t0 = time.time()
        rounds = 10
        for i in range(1, rounds+1):
            print('start round %s'% i)
            for h in hndles:
                if not h: continue
                result = natlinkutilsqh.SetForegroundWindow(h)
                if result:
                    self.assert_not_equal(unknownHndle, h, "hndle should not be unknown (fake) hndle")
                else:
                    self.assert_equal(unknownHndle, h, "hndle should be one of the valid hndles")

            time.sleep(0.1)
        t1 = time.time()
        deliberate = rounds*0.1
        total = t1 - t0
        nettime = t1 - t0 - deliberate
        nswitches = rounds*len(hndles)
        timeperswitch = nettime/nswitches
        print('nswitches: %s, nettime: %.4f, time per switch: %.4f'% ( nswitches, nettime, timeperswitch))
        print('total: %s, deliberate: %s'% (total, deliberate))

           
    def log(self, t):
        if self.isConnected:
            natlink.displayText(t, 0)
        else:
            print(t)
        print(t, file=open(logFileName, "a"))

def run():
    print('starting UnittestClipboard') 
    # trick: if you only want one or two tests to perform, change
    # the test names to her example def test....
    # and change the word 'test' into 'tttest'...
    # do not forget to change back and do all the tests when you are done.
    # each test does a natConnect and a natDisconnect...
    sys.stdout = open(logFileName, 'a')
    sys.stderr = open(logFileName, 'a')
    
    suite = unittest.makeSuite(UnittestClipboard, 'test')
    result = unittest.TextTestRunner().run(suite)
    
if __name__ == "__main__":
    print("run, result will be in %s"% logFileName)
    run()



#!/usr/bin/env python
import re

UdoMoves = re.compile(r'[A-Z]\d*')

UdoDict = dict(D='down', U='up', L='left', R='right')

def UdoToUnimacro(s):
    """convert D3 to {down 3} etc
    
>>> UdoToUnimacro('abcd3')
'abcd3'
>>> UdoToUnimacro('L1D3')
'{left}{down 3}'
>>> UdoToUnimacro('L3;;;U1;R123')
'{left 3};;;{up};{right 123}'
    
    """
    if UdoMoves.search(s):
        s = UdoMoves.sub(UdoSub, s)
    return s

def UdoSub(pat):
    """return substituted pattern of a part of the match
    
    """
    text = pat.group()
    if len(text) == 0:
        return text
    elif len(text) == 1:
        direction = text
        count = 1
    else:
        direction, count = text[0], int(text[1:])
    if direction in UdoDict:
        d = UdoDict[direction]
        if count == 1:
            return '{%s}'% d
        else:
            return '{%s %s}'% (d, count)
    return pat

def _test():
    import doctest
    return doctest.testmod()

if __name__ == "__main__":
    _test()



# This file was part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see https://qh.antenna.nl/unimacro/aboutunimacro/index.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory 
# Now in the dictation-toolbox: https://github.com/dictation-toolbox/unimacro

# _lines.py 
#  written by: Quintijn Hoogenboom (QH software, training & advies)
#  August 2003
#  add column: 23/4/2018
#  get it working for python3, 2022 (QH)
#
#pylint:disable=R0904, R0912, C0209, R0915, R0913

"""lines can be selected, with or without line number, copied, moved etc

In the second part of the command, also be triggered by "THAT", an appropriate action can be performed.

Examples: "Line copy", "Line 34", "Line 343 select", "Line 2 Through 5", 
"Line 15 Plus 3 Duplicate",  etc

More elaborate actions are: "Line 3 Move To 6", "Line 3 Through 7 Copy To 10", "Line 5 Move Up 2", "Line 10 Plus 2 Copy Down 10".

If you want to separate the selection and the action you can first say for example: "Line 11 Through 15" and as action "That Copy Up 3". 

If you use the Through part with a lower number, the next higher possible number is calculated, for example "Line 23 Through 5" means "Line 23 Through 25" and
"Line 98 Through 23" means "Line 98 Through 123".


All selection and action stuff is performed in the gotResults part of the grammar.  The following instance variables are therefore maintained: 
self.firstPart = 'that' | 'line' | 'linesnum'
self.action = '' | string (simple action) |
                   list: ['move'|'copy' , 'to'|'up'|'down']
self.numlines = number of lines to select - 1
self.line = linenumber to select
self.movecopyto = line number or number of lines to do the move copy action to
self.previouslines = Previous is used in the command for line selection (0|1)
self.nextLines = Next is used in the command for line selection (0|1)
lines canIn order to collect the numbers in the grammar correctly,
The numbers are collected by the <integer> rule, and subrules, see in natlinkutilsbj.

self.lastDirection gives the last selection direction, persistent

For more information on this number part, see grammar _testnumbersspokenforms.py
"""
import natlink
from natlinkcore import natlinkutils
import unimacro.natlinkutilsbj as natbj
from dtactions.unimacro import inivars
from dtactions.unimacro import unimacroutils
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.sendkeys import sendkeys as keystroke
from dtactions.unimacro import unimacroactions as actions

class LinesError(Exception):
    pass

counts = list(range(1,20)) + list(range(20,50,5)) + list(range(50,100,10)) + list(range(100, 1001, 100))
#print 'counts: %s'% counts

ancestor = natbj.DocstringGrammar
class ThisGrammar(ancestor):
    language = unimacroutils.getLanguage()        
    iniIgnoreGrammarLists = ['count', 'taskcount', 'taskapplication'] # are set in this module
                                                # taskcount and taskapplication only in very special
                                                # case, see Arnoud...
    if language == "nld":
        name = "regels"
        numGram = natbj.numberGrammarTill999['nld']
    else:
        name = "lines"
        numGram = natbj.numberGrammarTill999['enx']
    gramSpec = """
<before> = Here;
<linesthat> exported = (THAT | <lines> | <paras>) <action>;
<linesnum> exported = <linenum> | <linenum> <action>;
<lines> = line | this line | <before> line | 
           {count} lines| <before> {count} lines |
           these {count} lines|
           (previous|next) (line | {count} lines);
<paras> = para | this para | <before> para | <before> this para |
           {count} paras | <before> {count} paras |
           these {count} paras|
           (previous|next) (para | {count} paras);
<linenum> = line (<integer> | <integer>(through<integer> | plus {count}) | back);
<action> = <column> | {simpleaction} | <movecopyaction>;
<column> = column <integer>;
<movecopyaction> = (move | copy) (((down|up) {count})|(to <integer>));
"""+numGram


    def initialize(self):
        if not self.language:
            return
        
        self.load(self.gramSpec)
        self.setNumbersList('count', counts)
        
        # only for coupling with tasks search (Arnoud)
        if self.enableSearchCommands:
            self.setNumbersList('taskcount', self.tasksGrammar.taskCounts)
            self.setList('taskapplication', self.tasksGrammar.ini.get('application'))
        
        self.switchOnOrOff()
        self.mode = 'inactive'
        self.lastDirection = 'down'

    def gotBegin(self,moduleInfo):
        if self.checkForChanges:
            self.checkInifile()
        if self.prevModInfo == moduleInfo:
            return
        # changing sets, as moduleInfo changed
##        if not moduleInfo[0]: # no valid module info, do nothing
##            print 'no module info, skipping'
##            return
        # if window changes reset base:
        self.maxBase = 0
        self.base = 0

        self.progInfo = unimacroutils.getProgInfo(modInfo=moduleInfo)
        if unimacroutils.matchWindow(self.ignore, progInfo=self.progInfo):
##            print 'progInfo in ignore, skipping: %s'% self.ignore
            return
        if self.windowPolicy(moduleInfo, self.progInfo):
            if self.mode != 'active':
                self.activateAll()
                self.mode = 'active'
            else:
                pass
##                print '%s, remains active'% self.getName()
        else:
            if self.mode != 'inactive':
##                print '%s: non matching window deactivate: %s'% \
##                      (self.getName(), moduleInfo[1])
                self.deactivateAll()
                self.mode = 'inactive'
            else:
                pass
##                print '%s, remains inactive'% self.getName()
        self.prevModInfo = moduleInfo
        # self.progInfo = progInfo

    def gotResultsInit(self,words,fullResults):
        self.currentLine = None
        self.previousLines = self.nextLines = 0
        self.firstPart = ''        
        self.line = 0
        self.through = 0
        self.numlines = 1 # for selecting line number plus count
        self.movecopyto = 0 # line number or number of lines to move or copy to
        self.action = None # for simple actions the action string or list
                           # for movecopy actions set to ['move','to'], ['copy', 'up'] etc
        self.paras = False  # normally lines
        self.column = 0
        self.resetWordVariables()
        
        
        
    def resetWordVariables(self):
        """for the processing of the word copy paste commands
        """
        # self.count is set in the wordspec rule and
        # self.wordAction is set in the rules wordspecifyaction and afterwordoptional.
        # self.direction is default 'right', but can be adapted to 'left' in wordspec rule.
        # in doWordAction, the action is only performed if variable have been set.
        # special case: self.count = 0, only click, in afterwordoptional.
        self.wordAction = None
        self.searchAction = None
        self.count = None
        self.direction = 'right'
        
    ## words:
    # word-action handling:
    def rule_words(self, words):
        """# rules to manage words (copy/paste etc)
           # optional part after a cut/copy action, to move mouse
           # to a new location and do a paste there:
           <wordspecifyaction> | <before> <wordspecifyaction> |
           <wordspecifyaction> <afterwordoptional> | <before> <wordspecifyaction> <afterwordoptional>
        """
        print(f'never comes here! {words}')

    
    def subrule_wordspec(self, words):
        """word | {n2-20} words | word (left|right) | {n2-20} words (left|right)
        """
        lenwords = len(words)
        print(f'wordspec, got: {words} ({lenwords})')
        for i, w in enumerate(words):
            if self.hasCommon(w, 'word'):
                self.count = 1
                if i < lenwords - 1:
                    if self.hasCommon(words[i+1], 'left'):
                        direction = 'left'
            elif self.hasCommon(w, 'words'):
                if not self.count:
                    print('error in grammar lines, wordspec, count should have a value by now: %s\nprocessing word: %s (words: %s)'% \
                          (self.count, w, words))
                if i < lenwords - 1:
                    if self.hasCommon(words[i+1], 'left'):
                        self.direction = 'left'
            elif self.hasCommon(w, ['left', 'right']):
                continue # should have been processed already
            else:
                self.count = self.getNumberFromSpoken(w)
                if not self.count:
                    print('error in grammar lines, wordspec, count been caught here: %s\nprocessing word: %s (words: %s)'% \
                          (self.count, w, words))
                continue
            # process the count:
            if self.wordAction:
                self.doWordAction() # and flush
            if self.searchAction:
                self.doSearchAction()  # and flush

    def subrule_wordspecifyaction(self, words):
        """<wordspec> <wordaction> |
               <wordaction> <wordspec>
        """
        print("should never come in subrule_wordspecifyaction %s"% words)
        
        
    def subrule_wordaction(self, words):
        """{wordaction}| (search ({taskcount}|{taskapplication}))"""
        
        # wordactions, simple actions from inifile
        # search, coupling with tasks grammar for search command after a word
        # was selected, for Arnoud van den Eerenbeemt, QH august 2011
        
        self.wordAction = self.getFromInifile(words, 'wordaction')
        if self.wordAction:
            if self.count:
                self.doWordAction() # only if data
        else:
            if self.hasCommon(words, 'search'):
                self.searchAction = ['search', words[-1]]
                if self.count:
                    self.doSearchAction()

    def subrule_afterwordoptional(self, words):
        """{afterwordaction} | {afterwordaction} <wordspec> | <wordspec> {afterwordaction}
        """
        print("afterwordoptional, got: %s"% words)
        self.wordAction = self.getFromInifile(words, 'afterwordaction')
        if not self.doMouseMoveStopClick():
            self.count = None
            self.wordAction = 0
            print('aborted doMouseMoveStop action, stop this rule')
            return
        # because <wordspec> is optional, test here if it is yet to come (nextRule)
        if self.count:
            # had wordspec rule:
            self.doWordAction()
        elif self.count is None and self.nextRule != 'wordspec':
            print('no word specification in afterwordoptional')
            self.count = 0 # go with a single click if no wordspec has been given
            self.doWordAction()
            
        
    def doMouseMoveStopClick(self):
        """wait for mouse starting to move, stopping and then click
        """
        action("ALERT")
        if not action("WAITMOUSEMOVE"):
            action("ALERT 2")
            return None
        if not action("WAITMOUSESTOP"):
            return None
        action("ButtonClick")
        action("ALERT")
        return 1

    def doWordAction(self):
        """process count, direction and action
        
        if done, reset variables,
        if variables missing, do nothing
        """
        if self.count is None or self.wordAction is None:
            print('not ready for word action: %s, %s, %s'% (self.count, self.direction, self.wordAction))
            return
        if self.count == 0:
            #print 'doWordAction, single click: %s, %s, %s'% (self.count, self.direction, self.wordAction)
            pass
        else:
            #print 'doWordAction (select words): %s, %s, %s'% (self.count, self.direction, self.wordAction)
            wordSelect = "SELECTWORD %s, %s"% (self.count, self.direction)
            action(wordSelect)
        if self.wordAction:
            action(self.wordAction)
        self.resetWordVariables()

    def doSearchAction(self):
        """process count, direction and action, in this case a search into another window
        
        (coupling with tasks grammar)
        
        if done, reset variables,
        if variables missing, do nothing
        """
        if self.count is None or self.searchAction is None:
            print('not ready for word action: %s, %s, %s'% (self.count, self.direction, self.wordAction))
            return
        if self.count == 0:
            #print 'doWordAction, single click: %s, %s, %s'% (self.count, self.direction, self.wordAction)
            pass
        else:
            #print 'doWordAction (select words): %s, %s, %s'% (self.count, self.direction, self.wordAction)
            wordSelect = "SELECTWORD %s, %s"% (self.count, self.direction)
            action(wordSelect)
        if self.searchAction:
            print('now the search action in the tasks grammar: %s'% self.searchAction)
            self.tasksGrammar.rule_searchinothertask(self.searchAction)
        self.resetWordVariables()


    def gotResults_linenum(self,words,fullResults):
        """starting a linenum rule
        
        when the command is line back, this is intercepted immediately
        otherwise the waiting for a number is started
        """
        if self.hasCommon(words, 'back'):
            self.firstPart = 'lineback'           
            return
        self.firstPart = 'linesnum'
        self.lastDirection = 'down'
        self.collectNumber()
        if self.hasCommon(words, ['line']):
            self.waitForNumber('line')
        elif self.hasCommon(words, ['through']):
            self.waitForNumber('through')
        elif self.hasCommon(words, ['plus']):
            gotCounts = self.getNumbersFromSpoken(words, counts)
            if gotCounts:
                self.numlines = gotCounts[-1] + 1
            

    def gotResults_that(self,words,fullResults):
        self.firstPart = 'that'

    def gotResults_lines(self,words,fullResults):
        self.lastDirection = 'down'
        self.firstPart = 'lines'
        if self.hasCommon(words, ['previous']):
            self.previousLines = 1
            self.lastDirection = 'up'
        if self.hasCommon(words, ['next']):
            self.nextLines = 1
        if self.hasCommon(words, ['line']):
            self.numlines = 1
        if self.hasCommon(words, ['lines']):
            getCounts = self.getNumbersFromSpoken(words, counts)
            if getCounts:
                if len(getCounts) > 1:
                    print('warning, more counts found: %s (take the first)'% getCounts)
                self.numlines = getCounts[0]
            else:
                print('should collect a count, nothing found, take 1')

    def gotResults_paras(self,words,fullResults):
        self.lastDirection = 'down'
        self.firstPart = 'paras'
        if self.hasCommon(words, ['previous']):
            self.previousLines = 1
            self.lastDirection = 'up'
        if self.hasCommon(words, ['next']):
            self.nextLines = 1
        if self.hasCommon(words, ['para']):
            self.numparas = 1
        if self.hasCommon(words, ['paras']):
            getCounts = self.getNumbersFromSpoken(words, counts)
            if getCounts:
                if len(getCounts) > 1:
                    print('warning, more counts found: %s (take the first)'% getCounts)
                self.numlines = getCounts[0]
            else:
                print('should collect a count, nothing found, take 1')


    def gotResults_action(self,words,fullResults):
        self.action = self.getFromInifile(words, 'simpleaction')

    def gotResults_column(self,words,fullResults):
        """wait for integer, action in gotResults
        """
        self.collectNumber()
        self.waitForNumber('column')


    def gotResults_movecopyaction(self,words,fullResults):
        self.collectNumber()
        self.action = [None, None]
        if self.hasCommon(words, ['move']):
            self.action[0] = 'move'
        elif self.hasCommon(words, ['copy']):
            self.action[0] = 'copy'

        if self.hasCommon(words, ['down']):
            self.action[1] = 'down'
            self.movecopyto = self.getNumberFromSpoken(words[-1], counts)
        elif self.hasCommon(words, ['up']):
            self.action[1] = 'up'
            self.movecopyto = self.getNumberFromSpoken(words[-1], counts)
        elif self.hasCommon(words, ['to']):
            self.action[1] = 'to'
            self.waitForNumber('movecopyto')
        print('movecopyto: %s (%s)'% (self.movecopyto, type(self.movecopyto)))


    def gotResults_before(self,words,fullResults):
        if self.hasCommon(words, 'here'):
            natlinkutils.buttonClick('left', 1)
        
    def gotResults(self,words,fullResults):
        comment = 'command: %s'% ' '.join(words)
        self.prog = self.progInfo.prog
        self.collectNumber()
        if self.movecopyto and self.action[1] == 'to':
            if self.lineNumbersModuloHundred:
                self.movecopyto = self.convertLineNumberModulo(self.movecopyto)
            else:
                # absolute:
                self.movecopyto = int(self.movecopyto)
        if self.line:
            if self.lineNumbersModuloHundred:
                self.line = self.convertLineNumberModulo(self.line)
            else:
                # absolute:
                self.line = int(self.line)
            
        if self.through:
            intThrough = int(self.through)
            if intThrough > self.line:
                self.through = intThrough # always absolute
            else:
                
                if len(self.through) == 2:
                    modulo = 100
                else:
                    modulo = 10
                print('modulo for through: %s'% modulo)
                intThrough = getLineRelativeTo(intThrough, self.line, modulo=modulo,
                                               minLine=self.line)
                self.through = intThrough

            ## should not happen often:
            if self.line > self.through:
                # calculate the next higher number respective to self.line
                ndigits = len(repr(self.through))
                steps = 10**ndigits
                leftPart = self.line//steps
                newThrough = leftPart*steps + self.through
                while newThrough < self.line:
                    leftPart += 1
                    newThrough = leftPart*steps + self.through
                self.through = newThrough
                
            self.numlines = self.through - self.line + 1
        #print 'line: "%(line)s", to: "%(through)s", movecopyto: "%(movecopyto)s"' \
        #     ', numlines "%(numlines)s", action: "%(action)s"'%  self.__dict__

        # doing the selection part:
        T = []
        self.goingUp = 0
        if self.previousLines:
            self.goingUp = 1
        if self.nextLines:
            T.append('{extdown}')
        if self.firstPart == 'that':
            pass
        elif self.firstPart == 'lineback':
            T.append('<<lineback>>')
        elif self.firstPart == 'lines':
            if self.goingUp:
                T.append('<<selectpreviousline>>')
                if self.numlines > 1:
                    T.append('<<selectup %s>>' % (self.numlines-1,))
            else:
                T.append('<<selectline>>')
                if self.numlines > 1:
                    T.append('<<selectdown %s>>' % (self.numlines-1,))
        elif self.firstPart == 'paras':
            print('paragraphs!')
            if self.goingUp:
                print('vorige alinea')
                T.append('<<selectpreviouspara>>')
                if self.numlines > 1:
                    T.append('<<selectparaup %s>>' % (self.numlines-1,))
            else:
                print('deze of volgende alinea')
                T.append('<<selectpara>>')
                if self.numlines > 1:
                    T.append('<<selectparadown %s>>' % (self.numlines-1,))
            
        elif self.firstPart == 'linesnum':
            T.append('<<gotoline %s>>'% self.line)
            if self.numlines > 1:
                T.append('<<selectline>>')
                T.append('<<selectdown %s>>' % (self.numlines-1,))
            elif self.action is not None:
                # only when you call for a single line without action, NO
                # selection is done
                T.append('<<selectline>>')

        # t1 = time.time ()
        action(''.join(T), comment=comment)
        # t2 = time.time ()
##        print 'line select action: %s'% (t2-t1)
        T = []
        
        # doing the action part:
        if self.column:
            keystroke("{right %s}"% self.column)
            return
        if not self.action:
            return
        if isinstance(self.action, str):
            action(self.action, comment=comment)
        elif isinstance(self.action, list):
            if self.action[0] == 'move':
                T.append('<<cut>>')
            elif self.action[0] == 'copy':
                T.append('<<copy>>')
                if not self.prog in ['excel']:
                    if self.lastDirection == 'up':
                        T.append('{extleft}')
                    elif self.lastDirection == 'down':
                        T.append('{extright}')
            else:
                raise LinesError('invalid movecopy action (first word): %s'% self.action)

            print('gotResult: movecopyto: %s (%s)'% (self.movecopyto, type(self.movecopyto)))
            print('gotResult: numlines: %s'% self.numlines)

            if self.action[1] == 'up':
                if self.prog in ['excel']:
                    T.append("<<movetotopofselection>>")
                if self.prog in ['pycharm64', 'pycharm32']:
                    print('pycharm, %s'% repr(T))
                    if self.action[0] == 'copy':
                        T.append("{ctrl+d}")   # duplicate
                    T.append("{shift+ctrl+up %s}"% self.numlines)
                else:
                    # all other applications
                    T.append('{up %s}' % self.movecopyto)

            elif self.action[1] == 'down':
                if self.prog in ['excel']:
                    T.append("<<movetobottomofselection>>")
                    self.movecopyto += 1
                if self.prog in ['pycharm64', 'pycharm32']:
                    print('pycharm, %s'% repr(T))
                    if T and T[0] == "<<copy>>":
                        T = ["{ctrl+d}"]  ## duplicate
                    else:
                        T = []
                    T.append("{shift+ctrl+down %s}"% self.numlines)
                else:
                    # all other applications:
                    T.append('{down %s}' % self.movecopyto)
            elif self.action[1] == 'to':
                if self.action[0] == 'move' and not self.prog in ['excel']:
                    # if new line below the current line, the cut lines have to be accounted for (not in excel though)
                    currentLine = self.line or self.currentLine
                    if currentLine:
                        if self.movecopyto > currentLine + self.numlines:
                            self.movecopyto -= self.numlines
                    else:
                        print('_lines: movecopyaction: current line not known, move can be wrong if target is below current line')

                T.append('<<gotoline %s>>' % self.movecopyto)
            else:
                raise LinesError('invalid movecopy action (second word): %s'% self.action)
            T.append('<<movecopypaste>>')
            if self.numlines and not self.prog in ['excel']:
                T.append('{extup %s}{exthome}'% self.numlines)
##            action('<<upafterpaste>>', comment=comment)
            T.append('<<afterlines>>')
        if T:
            # t1 = time.time ()
            action(''.join(T), comment=comment)
            # t2 = time.time ()
##            print 'line action action: %s'% (t2-t1)

                
    def convertLineNumberModulo(self, num_as_string):
        """given the number as string, see if it is relative and convert to absolute number
        
        the value should be less than 100 or be prefixed with "0", and
        self.lineNumbersModuloHundred should be True, and
        the foreground app should be able to return the current line number
                (presently for Komodo, Excel and Visual Studio ("code"))
                
        When currentline > 100, then "07" should be treated as relative...
        """
        intLine = int(num_as_string)
        if not self.lineNumbersModuloHundred:
            # should not be here
            return intLine
        
        if intLine >= 100 or (intLine >= 10 and num_as_string.startswith('0')) or num_as_string.startswith('00'):
            return intLine # always absolute
        if not self.currentLine:
            self.app = actions.get_instance_from_progInfo(self.progInfo)
            if self.app:
                self.currentLine = self.app.getCurrentLineNumber()
            print(f'get current line number currentLine: {self.currentLine}')
        if num_as_string.startswith('0'):
            modulo = 100
        elif intLine < 10:
            modulo = 10
        else:
            modulo = 100
        if self.currentLine and self.currentLine > 99:
            # modulo trick only when you are above line 100
            intLine = getLineRelativeTo(intLine, self.currentLine, modulo=modulo)
            print(f'absolute: {intLine}, current: {self.currentLine}, modulo: {modulo}')
        return intLine

    def fillDefaultInifile(self, ini=None):
        """initialize as a starting example the ini file

        """
        ini = ini or self.ini
        ancestor.fillDefaultInifile(self, ini)
        if self.language == 'nld':
            self.ini.set('simpleaction', 'selecteren', '')
            self.ini.set('simpleaction', 'verwijderen', '{del}')
            self.ini.set('simpleaction', 'knippen', '{ctrl+x}')
            self.ini.set('simpleaction', 'kopiejeren', '{ctrl+c}')
            self.ini.set('simpleaction', 'dupliceren', '{ctrl+c}{right}{ctrl+v}')
            self.ini.set('simpleaction', 'wissen', '{del}')
            self.ini.set('simpleaction', 'einde', '<<endafterselection>>')
            self.ini.set('simpleaction', 'over plakken', '{ctrl+v}')
            self.ini.set('simpleaction', 'plakken', '<<realhome>>{ctrl+v}')
            self.ini.set('simpleaction', 'over plakken', '{ctrl+v}')
            self.ini.set('simpleaction', 'over plakken', '{ctrl+v}')
            self.ini.set('general', 'deactivate', 'sol')
            self.ini.set('general', 'ignore', 'empty')
            
        elif self.language == 'enx':
            self.ini.set('simpleaction', 'select', '')
            self.ini.set('simpleaction', 'delete', '{del}')
            self.ini.set('simpleaction', 'cut', '{ctrl+x}')
            self.ini.set('simpleaction', 'copy', '{ctrl+c}')
            self.ini.set('simpleaction', 'duplicate', '{ctrl+c}{right}{ctrl+v}')
            self.ini.set('simpleaction', 'end', '<<endafterselection>>')
            self.ini.set('simpleaction', 'paste', '<<realhome>>{ctrl+v}')
            self.ini.set('simpleaction', 'paste over', '{ctrl+v}')
            self.ini.set('general', 'deactivate', 'sol')
            self.ini.set('general', 'ignore', 'empty')
        else:
            self.ini.set('simpleaction', 'select', '')
            self.ini.set('general', 'ignore', 'empty')

    def fillInstanceVariables(self, ini=None):
        """fills instance variables with data from inifile

        overload for grammar lines: get activate/deactivate windows

        """
        try:
            #print 'fillInstantVariables for %s'% self
            ini = ini or self.ini
            self.lineNumbersModuloHundred = self.ini.getBool('general', 'line numbers modulo hundred')
            if self.lineNumbersModuloHundred:
                print('_lines: do "line numbers modulo hundred" if the application allows this')

            self.activateRules = ini.getDict('general', 'activate')
            if not self.activateRules:
                self.activateRules = {'all': None}
            if '*' in self.activateRules:
                del self.activateRules['*']
                self.activateRules['all'] = None
            for prog in self.activateRules:
                if self.activateRules[prog] == '*':
                    self.activateRules[prog]  = None
    ##        print 'self.activateRules: %s'% self.activateRules

            self.deactivateRules = ini.getDict('general', 'deactivate')
            if not self.deactivateRules:
                self.deactivateRules = {'none': None}
            if '*' in self.deactivateRules:
                del self.deactivateRules['*']
                self.deactivateRules['all'] = None
            for prog in self.deactivateRules:
                if self.deactivateRules[prog] == '*':
                    self.deactivateRules[prog]  = None
    ##        print 'self.deactivateRules: %s'% self.deactivateRules
            
            self.ignore = ini.getDict('general', 'ignore')
            if not self.ignore:
                try:
                    mgwn = {'nld': 'muisraster'}[self.language]
                except KeyError:
                    mgwn = 'mouse grid'
                self.ignore = {'empty': None, 'natspeak': mgwn}
            for prog in self.ignore:
                if self.ignore[prog] == '*':
                    self.ignore[prog]  = None
    ##        print 'self.ignore: %s'% self.ignore

            # Arnoud, same option needs to be set in _tasks
            self.tasksGrammarName = self.ini.get('general', 'enable search commands')
            self.tasksGrammar = self.enableSearchCommands = None
            # after a wordspec you can call "search number or task", numbers and tasks
            # corresponding with the _tasks grammar
            if self.tasksGrammarName:
                self.tasksGrammar = self.GetGrammarObject(self.tasksGrammarName)
                if self.tasksGrammar:
                    self.enableSearchCommands = 1
                    print('_lines, enable search commands, coupling grammar %s with %s'% (self.name, self.tasksGrammarName))
                else:
                    print('_lines, enable search commands failed, invalid name: %s'% self.tasksGrammarName)
                    self.enableSearchCommands = 0
            
        except inivars.IniError:
            print('IniError while initialising ini variables in _lines')

            
    def windowPolicy(self, modInfo, progInfo=None): 
        progInfo = progInfo or unimacroutils.getProgInfo(modInfo)
##        print 'window policy------progInfo: ', `progInfo`
        if unimacroutils.matchWindow(self.activateRules, progInfo=progInfo):
##            print 'matching activate: %s'% self.activateRules
            if not unimacroutils.matchWindow(self.deactivateRules, progInfo=progInfo):
                return 1
        return None
##        else:
##            print 'no positive match, deactivate:  %s'% self.activateRules
##        print 'window policy no match: %s'% modInfo[1]

def getLineRelativeTo(relativelinenum, currentLine, modulo=100, minLine=1, maxLine=None, visStart=None, visEnd=None):
    """return linenumber closest to currentLine, relative is modulo (normally 100)
    
    from lisp code of Mark, python (QH, modulolinenumbersmdl.py in miscqh) Oct 2013

    cannot below minLine (default 1)
    cannot above maxLine (if given)
    todo: if visStart and visEnd are given, do closest within the visible range...
    """
    n = relativelinenum
    if n < 0 or n >= modulo:
        raise ValueError("getLineRelativeTo, linenum must be between 0 and %s, not: %s"% (modulo, n))
    a = currentLine - currentLine%modulo + n
    if a < currentLine:
        b = a + modulo
    else:
        b = a - modulo
    if a < minLine:
        return b
    if b < minLine:
        return a
    if maxLine:
        if a > maxLine:
            return b
        if b > maxLine:
            return a

    if abs(a-currentLine) < abs(b-currentLine):
        return a
    return b

# standard stuff Joel (QH, Unimacro, python3):
try:
    thisGrammar
except NameError:
    thisGrammar = None

def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None 

if __name__ == "__main__":
    # here code to interactive run this module
    natlink.natConnect()
    try:
        thisGrammar = ThisGrammar(inifile_stem='_lines')
        thisGrammar.startInifile()
        thisGrammar.initialize()
        thisGrammar.progInfo = unimacroutils.getProgInfo()
        seqsAndRules = [(['line'], 'linenum'), (['seven', 'two', 'three'], '__0to9')]
        # ruleName: linenum, words: ['line'], FR: [('line', 'linenum'), ('seven', '__0to9'), ('two', '__0to9'), ('three', '__0to9')]
        # ruleName: __0to9, words: ['seven', 'two', 'three'], FR: [('line', 'linenum'), ('seven', '__0to9'), ('two', '__0to9'), ('three', '__0to9')]
        all_words = ['line', 'seven', 'two', 'three']
        FR = [('line', 'linenum'), ('seven', '__0to9'), ('two', '__0to9'), ('three', '__0to9')]
        thisGrammar.gotResultsInit(all_words, FR)
        thisGrammar.gotResults_linenum(['line'], FR)
        thisGrammar.gotResults___0to9(['seven', 'two', 'three'], FR)
        thisGrammar.gotResults(all_words, FR)
    finally:
        natlink.natDisconnect()
elif __name__.find('.') == -1:
    # called from the loader, when starting Dragon/Natlink:
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()



#
# frescobaldi.py
#
# Written by: Quintijn Hoogenboom (QH softwaretraining &
# advies),2002, revised October 2003
# march 2011: change sol to frescobaldi
"""grammar rules for frescobaldi (lilypond) music typesetting

The note input utterances, which are the most important, can contain
complete or incomplete notes and other parts (eg bar).

The cursor can be at several places, which is primary examined by getNextPrev, which
results in several instance variables.

- if text is selected:
-- the selection is replaced by the new input

- if the cursor is surrounded by whitespace (space of newline):
-- a complete note + what follows is entered new
-- an incomplete note is added to the previous note (eg a rythm or a height octave up/down correction)
-- other string parts are entered as new

- if the cursor is at the start of a note:
-- a complete or incomplete note changes the next note, and inserts what follows in the utterance
-- other string parts are inserted before the cursor

- if the cursor is just after a string,
-- to be filled in

- if the command says insert, a whitespace on both sides of the cursor or selection is ensured

The relevant information from the frescobaldi input window is retrieved by routines in
actionclasses/frescobaldi-actions.py.
The most important routines are:
- hasSelection, retrieving the selected text or returning None
- getPrevNext, retrieving 1 character on both sides of the cursor
- getNextLines(n=1): return n lines after the cursor
- getPrevLines(n=1): return n lines before the cursor

If just input notes are given, the strategy of insert/update of notes is
- cursor not just before another note or character (|):
  - if note is complete: insert
  - if note is partial (eg additional rythm or addition): update previous note
  
- cursor just before a note (or other character):
  - update the following note
  
"""

from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj
from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj
import natlink
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action
from itertools import cycle
from dtactions.unimacro import unimacroactions as actions
import re
import copy
import time
from lynote import LyNote, analyseString, getNotesFromWordsList, join

reRythmAttr =  re.compile(r"([132468.-]+)(.*)")


ancestor = natbj.DocstringGrammar
class ThisGrammar(ancestor):
    name = 'frescobaldi'

    def initialize(self):
        global language
        if self.load(self.gramSpec):
            print 'grammar frescobaldi (%s) active'% self.name
            self.prevHandle = -1
        else:
            language = ""
        self.frescobaldi = None
        self.rythmicPattern, self.rythmicKeys = None, None
                        
    def gotBegin(self, moduleInfo):
        winHandle = moduleInfo[2]
        if self.prevHandle == winHandle: return
        self.prevHandle = winHandle
        if moduleInfo[0].lower().find('frescobaldi.exe') > 0 and unimacroutils.isTopWindow(moduleInfo[2]):
            if self.frescobaldi is None:
                progInfo = unimacroutils.getProgInfo(moduleInfo)
                self.frescobaldi = actions.get_instance_from_progInfo(progInfo)
                if not self.frescobaldi:
                    raise Exception("frescobaldi, cannot get frescobaldi actions module")
                
            if self.checkForChanges:
                #print 'grammar frescobaldi (%s) checking the inifile'% self.name
                self.checkInifile()
            self.switchOnOrOff()
        elif self.isActive():
            self.deactivateAll()


    def gotResultsInit(self,words,fullResults):
        """initialize values
        """
        self.note = ""
        self.rythm = ""
        self.point = "" # for rythms, halving a note or rest
        self.noteaddition = "" # up, down, slur
        self.separator = "" # bar
        self.pitch = "" # up or down octave(s)
        self.wordsList, self.notesIndexes = [], [] # for notesnavigate
        self.afterNotesList = [] # list items after the selected notes
        self.notesWanted = 0 # also for notesnavigate...
        self.newNotes = []
        self.hadFigure = 0
        self.nextchars = None
        self.prevchars = None
        self.justInsert = 1
        t0 = time.time()
        self.hasSelection = self.frescobaldi.hasSelection()
        t1 = time.time()
        print 'hasSelection, (%.4f seconds): %s'% (t1-t0, self.hasSelection)
        self.justLeft = self.justRight = None # to be filled with getPrevNext()
        self.inMiddleOfWord = None
        if self.hasSelection:
            print 'hasSelection, selected text: %s'% unimacroutils.getClipboard()
        self.hasAllInfoNext = False ## set to true if nextlines end with double \n
        self.hasAllInfoPrev = False ## set to true if prevlines contains a double \n so starts with ir
        
    def rule_testgetprevnext(self, words):
        "get previous next"
        ## test just notes around the cursor:
        if self.hasSelection:
            print "selection: %s"% self.hasSelection
        else:
            t0 = time.time()
            self.getPrevNext()
            t1 = time.time()
            print '----testgetprevnext, time: %.4f'% (t1-t0,)
            print 'justLeft: %s (%s), justRight: %s (%s)'% (repr(self.justLeft), len(self.justLeft), repr(self.justRight), len(self.justRight))
            print 'inMiddleOfWord: %s'% self.inMiddleOfWord
            print 'hasWhiteSpaceAround: %s, atEndOfLine: %s'% (self.hasWhiteSpaceAround, self.atEndOfLine)
            print 'hasWhiteSpaceBefore: %s, hasWhiteSpaceAfter: %s'% (self.hasWhiteSpaceBefore, self.hasWhiteSpaceAfter)
            print '----'
            
    def rule_testgetnotes(self, words):
        """get (next|previous) (note | ({n2-10} notes))"""
        n = self.getNumberFromSpoken(words[-2]) or 1

        if self.hasCommon(words, 'next'):
            wordsList, notesIndexes, beforeString, afterString, afterNoteComplete = self.getNextNotes(n)
            print '--- next %s notes:'% n
            print 'before: %s, after: %s, notesfound: %s'% (repr(beforeString), repr(afterString), len(notesIndexes))
            print 'wordsList: %s'% wordsList
            print 'notesIndexes: %s'% notesIndexes
            print 'notes: %s'% repr([str(wordsList[i]) for i in notesIndexes])
        elif self.hasCommon(words, 'previous'):
            wordsList, notesIndexes, beforeString, afterString, beforeNoteComplete= self.getPrevNotes(n)
            print '--- previous %s notes:'% n
            print 'before: %s, after: %s, notesfound: %s'% (repr(beforeString), repr(afterString), len(notesIndexes))
            print 'wordsList: %s'% wordsList
            print 'notesIndexes: %s'% notesIndexes
            print 'notes: %s'% repr([str(wordsList[i]) for i in notesIndexes])
        else:
            print 'wrong word in grammar rule, not next or previous found: %s'% words
            
        
    def rule_sequence(self, words):
        "(<note>|<pitch>|<rythm>|<separator>|<noteaddition>)+"
        pass
    
    def subrule_note(self, words):
        "{notes} [{notesalteration}]"
        for w in words:
            note = self.getFromInifile(w, 'notes')
            if note:
                if self.note:
                    self.finishNote()
                self.note = note
                continue
            altnote = self.getFromInifile(w, 'notesalteration')
            if altnote:
                if not self.note:
                    raise ValueError("should have note before making a notesalteration: %s (%s)"% (altnote, w))
                self.note = self.noteAlteration(self.note, altnote)
    
    def noteAlteration(self, note, alteration):
        """make a note sharp or flat
        
        way to do depends on the inputlanguage of lilypond
        """
        return note + alteration

    def subrule_pitch(self, words):
        "{pitchs}"
        for w in words:
            pitch = self.getFromInifile(w, 'pitchs')
            if pitch:
                self.pitch += pitch
                continue                                

    def subrule_noteaddition(self, words):
        "{noteadditions}"
        for w in words:
            self.noteaddition = self.getFromInifile(w, 'noteadditions')
        
    def subrule_rythm(self, words):
        "{rythms}|{rythms} point"
        # switch off rythmicPattern as soon as a rythm is called:
        # if rythm contains a space, it is converted into a rythmicPattern        
        self.rythmicPattern = None
        for w in words:
            if self.hasCommon(w, "point"):
                self.point += "."
            else:
                rythm = self.getFromInifile(w, 'rythms')
                if self.rythm and self.note:
                    self.finishNote()
                if rythm.find(' ') > 0:

                    self.rythmicPattern = cycle(self.analyseRythmicPattern(rythm))
                    print('self.rythmicPattern: %s'% self.rythmicPattern)
                else:
                    self.rythm = rythm

    def analyseRythmicPattern(self, rythm):
        """analyse rythmic pattern
        
        it consists of rythms, separated by spaces.
        a rythm can be augmented with noteadditions, like beams ([]) or slurs ()
        '-' as key is no new rythmic value
        
        return a list of 2-tuples [(rythm, attr), (rythm, attr)]
        rythm '8.( 16) 8' returns [('8.', '('), ('16',')', ('8', '')]
        """
        if not ' ' in rythm:
            raise ValueError('analyseRythmicPattern, no space found in pattern: %s'% rythm)
        rythms = [r.strip() for r in rythm.split()]
        print("rythm: %s, rythms: %s"% (rythm, rythms))
        L = []
        for r in rythms:
            print('rythmic part: %s'% r)
            m = reRythmAttr.match(r)
            if m:
                L.append(m.groups())
            else:
                raise ValueError("invalid rythm in part %s of rythmic pattern: %s"% (r, rythm))
        
        print("rythmic L: %s"% L)
        return L

    def subrule_separator(self, words):
        "{separators}"
        self.finishNote()
        s = []
        for w in words:
            separator = self.getFromInifile(w, 'separators')
            self.newNotes.append(separator)
            
    def rule_figure(self, words):
        """here figure {figures}+"""
        # start with the pointer in the music pane
        # always click
        unimacroutils.buttonClick()
        unimacroutils.visibleWait()
        self.hadFigure = 1

        selection = self.getFigurePart()
        if not selection:
            print 'rule_figure, no selection found, stop macro'
            return
        
        # print 'figurepart: %s'% selection
        L = []
        previous = None
        for w in words:
            figure = self.getFromInifile(w, 'figures')
            # print 'figure: %s, result: %s'% (w, figure)
            if figure:
                if previous and self.isFigure(previous) and not self.isFigure(figure):
                    if figure.startswith("_"):
                        figure = figure.lstrip("_")
                    L[-1] = L[-1] + figure
                else:
                    L.append(figure)
                previous = figure
        self.keystroke("<" + ' '.join(L) + ">")  # allow for { , not being a control character!
        
    def isFigure(self, figure):
        """return True if figure is a cipher, 1, 2, ...
        """
        try:
            int(figure)
        except ValueError:
            return False
        return True
            
    def rule_notesaddition(self, words):
        "<notesnavigate>{notesaddition}"
        # if words end with Off, On or End, they are in front of the first word and separated from the last.
        self.justInsert = False
        if not self.notesIndexes:
            print 'notesaddition, no notes found in selection'
            return
        
        prepost = self.getFromInifile(words[-1], 'notesaddition')
        print 'prepost: "%s"'% prepost
        doEachNote = False
        # to be switched on with eg ()() (eachnoteslur)
        if not prepost:
            return
        if ":" in prepost:
            pre, post = prepost.split(":")
        else:
            lenpart = len(prepost)/2
            pre, post = prepost[:lenpart], prepost[-lenpart:]
        if pre == post and len(pre) > 1:
            lenpart = len(pre)/2
            pre, post = pre[:lenpart], post[-lenpart:]
            print 'pre: %s, post: %s, do continuous'% (pre, post)
            doEachNote = True
        startNote = self.wordsList[self.notesIndexes[0]]
        print 'notesIndexes: %s, notesWanted: %s'% (self.notesIndexes, self.notesWanted)
        endNote = self.wordsList[self.notesIndexes[self.notesWanted-1]]
        
        startNote.updateNote(pre)
        endNote.updateNote(post)
        if doEachNote and self.notesWanted > 2:
            for i in range(1, self.notesWanted-1):
                print 'do extra for note %s'%i
                note = self.wordsList[self.notesIndexes[i]]
                note.updateNote(post)
                note.updateNote(pre)
                
        #for ending in ['On', 'Off']:  # strange bit here
        #    if post.endswith(ending):
        #        self.wordsList.insert(endNote, post)
        #        break
        #else:
        #    self.wordsList.insert(endNote+1, post)
        #
        #for ending in ['On', 'Off', 'End']:
        #    if pre.endswith(ending):
        #        self.wordsList.insert(startNote, pre)
        #        print 'pre "%s" ending: "%s", wordsList: %s'% (pre, ending, self.wordsList)
        #        break
        #else:
        #    print 'pre "%s" general, wordList: %s'% (pre, self.wordsList)
        #    self.wordsList.insert(startNote+1, pre)

        snippet = " ".join([str(x).strip() for x in self.wordsList])
        print 'new snippet: %s'% snippet
        keystroke(snippet)
        self.gotoNextNote()
        
    def rule_notechangemove(self, words):
        """<notesnavigate>(<sequence>|{noteaction}|forward)"""
        self.justInsert = False

        act =  self.getFromInifile(words[-1], 'noteaction')
        if act:
            action(act)
            return
        if self.hasCommon(words[-1], 'forward'):
            self.gotoNextNote()

    def rule_notesinsert(self, words):
        """[here] insert [<sequence>]"""
        ###
        ### position at correct position, possibly put a space.
        if self.hasCommon(words, 'here'):
            unimacroutils.buttonClick()
        self.getPrevNext()
        # position at left of word
        while self.inMiddleOfWord:
            keystroke("{left}")
            self.getPrevNext()
        # and put a space:
        keystroke("{space}{left}")
        self.getPrevNext()
        
      
     # def rule_test(self, words):
    #     """test position {direction}"""
    #     # position first at word boundary
    #     direction = self.getFromInifile(words[-1], 'direction')
    #     self.positionAtWordBoundary(direction)
    
    def subrule_notesnavigate(self, words):
        "here | [here] [next] (note | {n2-20} notes)"
        # leave next|previous for the moment, assume always count from the beginning      
        DIR = 'right'
        if self.hasCommon(words, 'here'):
            unimacroutils.buttonClick()
        try:
            nStr = self.getNumbersFromSpoken(words)[0] # returns a string or None
            n = int(nStr)
        except IndexError:
            n = 1
        if n == 1:
            gotoNext = self.hasCommon(words, 'next')
            if gotoNext:
                print 'move right one note! Todo'
        self.notesWanted = n
        print 'get %s notes'% n
        self.getNextNotes(n)
        print 'notesNavigate, wordsList: %s'% self.wordsList
        
       
    def getNextNotes(self, n=1):
        if self.nextchars is None:
            self.nextchars = self.frescobaldi.getNextLine()
        print 'nextchars: |%s|'% self.nextchars
        if self.justLeft is None:
            self.getPrevNext()
        if self.inMiddleOfWord:
            print 'inMiddleOfWord %s'% self.inMiddleOfWord
            if not self.prevchars:
                self.prevchars = self.frescobaldi.getPrevLine()
            lastWord = self.prevchars.split()[-1]
            numcharsleft = len(lastWord)
            print 'go left %s characters'% numcharsleft
            keystroke("{left %s}"% numcharsleft)
            self.getPrevNext()
            self.nextchars = self.frescobaldi.getNextLine()

        print 'nextchars: %s'% repr(self.nextchars)
        wordsList, notesIndexes = analyseString(self.nextchars)
        print 'wordsList: %s, notesIndexes: %s'% (wordsList, notesIndexes)
        notesFound = len(notesIndexes)
        
        if notesFound <= n:
            moreLines = 5
            print 'get %s more lines'% moreLines
            nextchars = self.frescobaldi.getNextLine(moreLines)
            # nextcharsplit = self.nextchars.split('\n\n') # double empty line ends the search
            # if len(nextcharsplit) > 1:
            #     self.hasAllInfoNext = True
            # self.nextchars = nextcharsplit[0]
            if nextchars == self.nextchars:
                raise Exception("getNextNotes, not enough notes found: %s, wanted: %s"% (notesFound, n))
            self.nextchars = nextchars
            wordsList, notesIndexes = analyseString(self.nextchars)
            notesFound = len(notesIndexes)
            print '----nextchars: %s'% repr(self.nextchars)
            if notesFound < n:
                print 'notesFound: %s'% notesFound
                print 'wordsList: %s, notesIndexes: %s'% (wordsList, notesIndexes)
                raise Exception("too little notes caught in %s lines down: %s, wanted: %s"%
                                (moreLines, notesFound, n))
        else:
            self.hasAllInfoNext = True
        wordsList, notesIndexes, beforeNotes, afterNotes, afterNoteComplete = \
                              getNotesFromWordsList(n, wordsList, notesIndexes)
        afterNoteComplete =afterNoteComplete or self.hasAllInfoNext
        return wordsList, notesIndexes, beforeNotes, afterNotes, afterNoteComplete
        

    def getPrevNotes(self, n=1):
        if self.justLeft is None:
            self.getPrevNext()
        if self.prevchars is None:
            self.prevchars = self.frescobaldi.getPrevLine()
        if self.inMiddleOfWord:
            print 'inMiddleOfWord %s'% self.inMiddleOfWord
            lastWord = self.prevchars.split()[-1]
            numcharsleft = len(lastWord)
            print 'go left %s characters'% numcharsleft
            keystroke("{left %s}"% numcharsleft)
            self.justLeft, self.justRight = self.prevchars[-numcharsleft-1], self.prevchars[-numcharsleft]
            print 'after go left: justLeft: %s, justRight: %s'% self.justLeft, self.justRight
            self.getPrevNextVariables(self, self.justLeft, self.justRight)
            self.prevchars = self.prevchars[:-numcharsleft]
            print 'prevchars after go left to beginning of note: %s'% repr(self.prevchars)
        
        print 'prevchars: %s'% repr(self.prevchars)
        wordsList, notesIndexes = analyseString(self.prevchars)
        print 'wordsList: %s, notesIndexes: %s'% (wordsList, notesIndexes)
        notesFound = len(notesIndexes)
        
        if notesFound <= n:
            moreLines = 5
            print 'get %s more lines'% moreLines
            prevchars = self.frescobaldi.getPrevLine(moreLines)
            if prevchars == self.prevchars:
                raise Exception("getPrevNotes, not enough notes found: %s, wanted: %s"% (notesFound, n))
            print '%s more lines prev:\n%s'% (moreLines, repr(prevchars))
            self.prevchars = prevchars

            wordsList, notesIndexes = analyseString(self.prevchars)
            notesFound = len(notesIndexes)
            print '----prevchars: %s'% repr(self.prevchars)
            if notesFound < n:
                print 'notesFound: %s'% notesFound
                print 'wordsList: %s, notesIndexes: %s'% (wordsList, notesIndexes)
                raise Exception("too little notes caught in %s lines up: %s, wanted: %s"%
                                (moreLines, notesFound, n))
        else:
            self.hasAllInfoPrev = True  ## know enough

        wordsList, notesIndexes, afterNotes, beforeNotes, beforeNoteComplete = \
                              getNotesFromWordsList(n, wordsList, notesIndexes, reverse=True)
        beforeNoteComplete = beforeNoteComplete or self.hasAllInfoPrev
        return wordsList, notesIndexes, afterNotes, beforeNotes, beforeNoteComplete
            

        
    def gotResults(self,words,fullResults):
        """flush last part of utterance"""
        if self.hadFigure:
            return
        
        output = []
        self.finishNote()
        print 'gotResults: newNotes: %s'% self.newNotes
        if not self.newNotes:
            return
        
        if self.hasSelection:
            keystroke("{del}")            
        
        if self.justLeft is None:
            self.getPrevNext()
        # print '------hasWhiteSpaceAround: %s, atEndOfLine: %s, hasWhiteSpaceBefore: %s'% (
            # self.hasWhiteSpaceAround, self.atEndOfLine, self.hasWhiteSpaceBefore)
        # print '===newNotes: %s'% self.newNotes
        
        noteOne = self.newNotes[0]
        if isinstance(noteOne, LyNote) and not noteOne.isNote():
                # update previous note, with partial note (eg a rythm)
                print 'noteOne is not a complete note: %s'% noteOne
                    # previous one:
                wordsList, notesIndexes, afterNotes, beforeNotes, beforeNoteComplete = self.getPrevNotes(1)
                lenBack = len(join(wordsList)) + len(afterNotes)
                keystroke("{shift+left %s}"% lenBack)
                wordsList[notesIndexes[0]].updateNote(noteOne)
                output.extend(wordsList)
                output.append(afterNotes)
                keystroke(join(output))
                output = []
                time.sleep(3)
                self.getPrevNext()
                if self.justRight:
                    print 'go one to the left'
                    keystroke(" {left}")
                    self.getPrevNext()
                self.newNotes.pop(0)
                
            
        # not a LyNote, such as a bar, put a space if it is not there
        if self.newNotes:
            noteOne = self.newNotes[0]
        if self.justLeft is None:
            self.getPrevNext()
            print("+++++hasWhiteSpaceBefore: |%s|"% self.hasWhiteSpaceBefore)
        if not self.hasWhiteSpaceBefore:
            print "insert space"
            output.insert(0, ' ')
            
        output.append(join(self.newNotes, separator=' '))
        if not (self.hasWhiteSpaceAfter or str(self.newNotes[-1]).endswith("{enter}")):
            output.append(' ')
        
        # print 'keystroking output: %s'% repr(output)
        keystroke(join(output))

        return
    
        print 'wordsList: %s'% self.wordsList
        if not self.wordsList:
            # check if just before a new note
            if self.justRight:
                print 'get next note'
                self.getNextNotes()

        if self.wordsList:
            print 'merge notes in wordsList'
            for i, note in enumerate(self.newNotes):
                if i < len(self.notesIndexes):
                    self.wordsList[self.notesIndexes[i]].updateNote(self.newNotes[i])
                else:
                    self.wordsList.append(note)
                    self.wordsList.append(' ')
            keystroke(join(wordsList))
            
            return
        # see if first note is a note, if so, output notes otherwise
        # change previous note (for example add rythm)
        
                
        
    def finishNote(self):
        """flush if far enough"""
        t = []
        noteDict = {}
        if self.note:
            if not self.rythm:
                if self.rythmicPattern:
                    rKey, rAttr = self.rythmicPattern.next()
                    if rKey == '-':
                        pass
                    else:
                        self.rythm = rKey
                    self.noteaddition = self.noteaddition + rAttr
                print 'rythm %s from pattern: %s'% (self.rythm, self.rythmicPattern)
        for name in ['note', 'pitch', 'rythm', 'point', 'noteaddition']:
            s = getattr(self, name, None)
            if s:
                # print 'name: %s, content: "%s"'% (name, s)
                t.append(s)
                setattr(self, name, "")

        noteText = ''.join(t)
        if noteText:
            note = LyNote(noteText)
            self.newNotes.append(note)
        print('newNotes: %s'% self.newNotes)

    def outputNewNotes(self):
        """just print the self.newNotes, separated by a space
        """
        output = []
        for note in self.newNotes:
            output.append(str(note))
            if not str(note).endswith("{enter}"):
                output.append(' ')
        self.keystroke(join(output))

    def getPrevNext(self):
        """get next en previous character
        
        assume there is no selection
        several instance variable are filled, for convenience
        self.justLeft, self.justRight : one character left and one char right of cursor
        self.hasWhiteSpaceAround
        self.hasWhiteSpaceBefore, self.hasWhiteSpaceAfter
        self.atEndOfLine
        """
        self.justLeft, self.justRight = self.frescobaldi.getPrevNext()
        # print 'from fr: |%s|(%s) |%s|(%s)'% (self.justLeft, len(self.justLeft), self.justRight, len(self.justRight))
        self.getPrevNextVariables(self.justLeft, self.justRight) 


    def getPrevNextVariables(self, justLeft, justRight):
        """compute several state variables, when justLeft and justRight are given
        
        mostly called from getPrevNext, but after a cursor movement justLeft and justRight are calculated,
        this routine calculates the remaining state variables.
        """
        self.atEndOfLine = (justRight == "\n")
        self.hasWhiteSpaceBefore = not justLeft.strip(' ')
        self.hasWhiteSpaceAfter = not justRight.strip(' ')
        self.inMiddleOfWord = not (self.hasWhiteSpaceAfter or self.hasWhiteSpaceBefore)
        self.hasWhiteSpaceAround = self.hasWhiteSpaceAfter and self.hasWhiteSpaceBefore
       
    def getFigurePart(self):
        # go left then right for selecting the note at the cursor position
        unimacroutils.saveClipboard()
        self.getPrevNext()
        if self.justRight != "<":
            # go left until < has been reached:
            for i in range(20):
                if self.getprevchar() == "<":
                    keystroke("{left}")
                    break
                keystroke("{left}")
            else:
                print 'getFigurePart, did not find "<"'
                keystroke("{right 20}")
                return
        # select step by step until > has been reached:
        for i in range(20):
            t = self.getselectionafterkeystroke("{shift+right}")
            if t and t.endswith(">"):
                break
        else:
            print 'getFigurePart, did not find ">"'
            unimacroutils.restoreClipboard()
            return
        result = t
        unimacroutils.restoreClipboard()
        return result
        
        
        
    def keystroke(self, keys):
        """internal keys for frescobaldi can have {] (braces), output them separate
        """
        keystroke(keys)
    
    def fillInstanceVariables(self):
        """fills the necessary instance variables
        
        configured from ini file, [general] section
        """
        self.newlineAfterBar = self.ini.getBool('general', 'newline after bar')

                
# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
thisGrammar = ThisGrammar()
if thisGrammar.gramSpec:
    thisGrammar.initialize()
else:
    thisGrammar = None

def unload():
    #pylint:disable=W0603
    global thisGrammar
    if thisGrammar: thisGrammar.unload()
    thisGrammar = None





#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# unittestNumbersConversion.py
#   This script performs some basic tests of the NatLink but
#   with the NumbersConversion subclass.
#   especially the numbers parts of a spoken form list
#
# run from a (preferably clean) US user profile, easiest from IDLE.
# do not run from pythonwin. 
#
import sys
import unittest
import types
import os
import os.path
import time
import traceback        # for printing exceptions
import TestCaseWithHelpers
import natlink
from natlinkcore import loader
from natlinkcore import natlinkstatus
from natlinkcore.gramparser import GrammarError, GrammarSyntaxError

status = natlinkstatus.NatlinkStatus()

from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action

class TestError(Exception):
    pass
ExitQuietly = 'ExitQuietly'

def getBaseFolder(globalsDict=None):
    """get the folder of the calling module.

    either sys.argv[0] (when run direct) or
    __file__, which can be empty. In that case take the working directory
    """
    globalsDictHere = globalsDict or globals()
    baseFolder = ""
    if globalsDictHere['__name__']  == "__main__":
        baseFolder = os.path.split(sys.argv[0])[0]
        print('baseFolder from argv: %s'% baseFolder)
    elif globalsDictHere['__file__']:
        baseFolder = os.path.split(globalsDictHere['__file__'])[0]
        print('baseFolder from __file__: %s'% baseFolder)
    if not baseFolder or baseFolder == '.':
        baseFolder = os.getcwd()
        print('baseFolder was empty, take wd: %s'% baseFolder)
    return baseFolder

thisDir = getBaseFolder(globals())

natconnectOption = 0 # or 1 for threading, 0 for not. Seems to make difference
                     # with spurious error (if set to 1), missing gotBegin and all that...
logFileName = os.path.join(thisDir, "testresult.txt")

#---------------------------------------------------------------------------
# These tests should be run after we call natConnect
# no reopen user at each test anymore..
# no default open window (open window will be the calling program)
# default .ini files pop up when you first run this test. just ignore them.
class UnittestNumbersConversion(TestCaseWithHelpers.TestCaseWithHelpers):
    def setUp(self):
        if not natlink.isNatSpeakRunning():
            raise TestError('NatSpeak is not currently running')
        self.connect()
        self.user = natlink.getCurrentUser()[0]
        self.setMicState = "off"

    def tearDown(self):
        try:
            # give message:
            self.setMicState = "off"
            # kill things
            self.clearTestFiles()
        finally:
            self.disconnect()

        
    def connect(self):
        # start with 1 for thread safety when run from pythonwin:
        natlink.natConnect(natconnectOption)

    def disconnect(self):
        natlink.natDisconnect()
        
    def log(self, t):
        # only log to file:
        log(t)

    def clearTestFiles(self):
        """remove .py and .pyc files from the natlinkmain test

        """
        return
        # try to remove __main__.ini, but gives unexpected windows popping up
        #userdir = status.getUnimacroUserDirectory()
        #language = status.getLanguage()
        #print 'userdir: %s, language: %s'% (userdir, language)
        #userinidir = os.path.join(userdir, '%s_inifiles'%language)
        #if os.path.isdir(userinidir):
        #    mainfile = os.path.join(userinidir, '__main__.ini')
        #    if os.path.isfile(mainfile):
        #        os.remove(mainfile)
        #else:
        #    raise OSError("clearTestFiles, should be a valid directory: %s"% userinidir)

    def wait(self, t=1):
        time.sleep(t)


    #---------------------------------------------------------------------------
    # This utility subroutine executes a Python command and makes sure that
    # an exception (of the expected type) is raised.  Otherwise a TestError
    # exception is raised

    def doTestForException(self, exceptionType,command,localVars={}):
        try:
            exec(command,globals(),localVars)
        except exceptionType:
            return
        raise TestError('Expecting an exception to be raised calling '+command)

    def doTestActiveRules(self, gram, expected):
        """gram must be a grammar instance, sort the rules to be expected and got
        """
        expected.sort()
        got = gram.activeRules
        got.sort()
        self.assertEqual(expected, got,
                         'Active rules not as expected:\nexpected: %s, got: %s'%
                         (expected, got))

    def doTestFuncReturn(self, expected,command,localVars=None):
        # account for different values in case of [None, 0] (wordFuncs)
        if localVars == None:
            actual = eval(command)
        else:
            actual = eval(command, globals(), localVars)

        if actual != expected:
            time.sleep(1)
        self.assertEqual(expected, actual, 'Function call "%s" returned unexpected result\nExpected: %s, got: %s'%
                          (command, expected, actual))
    
    def testGrammarNumberFunctionsWithoutNumbersIni(self):
        """test the functions that convert numbers lists
        
        when the numbers.ini has been removed (temporarily)
        """
        class TestGrammar(natbj.DocstringGrammar):
            name = "testgrammarone"
            def __init__(self):
                natbj.DocstringGrammar.__init__(self)
            def rule_one(self, words):
                'rule one'
                pass
       
        testGram = TestGrammar()
        
        expGramSpec =  '''<one> exported = rule one;'''
        self.assert_equal(expGramSpec, testGram.gramSpec,"gramspec not as expected")
        
        # test the numbers convert function
        numIni = testGram.getNumbersInifile()
        language = 'enx'
        self.assert_equal(language, testGram.language, "Testing should take place with 'enx' speech profile, not: %s"% testGram.language)
        hundredGot = numIni.getList(language + "_prefixes", "hundred")
        expHundredList = ["hundred", "one hundred", "a hundred"]
        self.assert_equal(expHundredList, hundredGot, "list of 'hundred prefixes should be as expected")
        result = testGram.getSpokenFormsAboveHundred(102, numIni)
        expResult = ["hundred two", "one hundred two", "a hundred two"]
        self.assert_equal(expResult, result, "list for 102 not as expected")
        result = testGram.getSpokenFormsAboveHundred(335, numIni)
        expResult = ['three hundred thirty-five', 'three thirty-five']
        self.assert_equal(expResult, result, "list for 335 not as expected")
    
        

def log(t):
    """log to print and file if present

    note print depends on the state of natlink: where is goes or disappears...
    I have no complete insight is this, but checking the logfile afterwards
    always works (QH)
    """
    print(t)
    if logFile:
        logFile.write(t + '\n')
    
#---------------------------------------------------------------------------
# run
#
# This is the main entry point.  It will connect to NatSpeak and perform
# a series of tests.  In the case of an error, it will cleanly disconnect
# from NatSpeak and print the exception information,
def dumpResult(testResult, logFile):
    """dump into 
    """
    if testResult.wasSuccessful():
        mes = "all succesful"
        logFile.write(mes)
        return
    logFile.write('\n--------------- errors -----------------\n')
    for case, tb in testResult.errors:
        logFile.write('\n---------- %s --------\n'% case)
        logFile.write(tb)
        
    logFile.write('\n--------------- failures -----------------\n')
    for case, tb in testResult.failures:
        logFile.write('\n---------- %s --------\n'% case)
        logFile.write(tb)

    


logFile = None

def run():
    global logFile, natconnectOption
    logFile = open(logFileName, "w")
    log("log messages to file: %s"% logFileName)
    log('starting unittestNatlink')
    # trick: if you only want one or two tests to perform, change
    # the test names to her example def test....
    # and change the word 'test' into 'tttest'...
    # do not forget to change back and do all the tests when you are done.
    suite = unittest.makeSuite(UnittestNumbersConversion, 'test')
##    natconnectOption = 0 # no threading has most chances to pass...
    log('\nstarting tests with threading: %s\n'% natconnectOption)
    result = unittest.TextTestRunner().run(suite)
    dumpResult(result, logFile=logFile)
    
    logFile.close()

if __name__ == "__main__":
    run()



#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
#   This script performs some basic tests of the NatLink system.  Dragon
#   NaturallySpeaking should be running with nothing in the editor window
#   (that you want to preserve) before these tests are run.
#   performed.
import sys
from pathlib import Path
thisDir = Path('.')
unimacroDir=(thisDir/'..').normPath()
if unimacroDir not in sys.path:
    print("add unimacroDir to sys.path: %s"% unimacroDir)
    sys.path.append(unimacroDir)
from dtactions.unimacro import unimacroutils
from natlinkcore import natlinkutils

import time
from dtactions.unimacro import unimacroactions as actions

import unittest
import UnimacroTestHelpers



##class UnimacroBasicTest(TestCaseWithHelpers.TestCaseWithHelpers):
class ClipboardTest(UnimacroTestHelpers.UnimacroTestHelpers):
      
    def setUp(self):
        actions.doAction("BRINGUP dragonpad; <<selectall>><<delete>>")

    def tearDown(self):
        actions.doAction("BRINGUP dragonpad; KW")
        
    def tttest_Something_in_unimacro(self):
        testWindowContents = self.doTestWindowContents
        actions.doKeystroke("testing")
        expected = "testing"
        testWindowContents(expected, "Something testing in ClipboardTest went wrong")
        # tearDown when DragonPad is already closed:
        actions.doAction("KW")
        print("test_Something_in_unimacro done")
 
    def tttest_Copy_and_paste_clipboard(self):
        testWindowContents = self.doTestWindowContents
        # This test handles several copy and paste meta actions.
        actions.doKeystroke("testing")
        actions.doAction("<<selectall>><<copy>>{ctrl+end}")
        actions.doAction("<<paste>>")
        # Note in DragonPad it apparently endswith "\r\n"...
        expected = "testingtesting\n"  # natqh getClipboard removes '\r', which was in natlink.getClipboard
        testWindowContents(expected, "Something copying and pasting went wrong")

    def test_Empty_clipboard(self):
        testWindowContents = self.doTestWindowContents
        # This test first empties the clipboard, copies with no selection on
##         and tests if the clipboard is empty.
##         If it is so it restores the previous clipboard and exits
        testActionResult = self.doTestActionResult
        actions.doKeystroke("testing")
        actions.doAction("<<selectall>>;<<copy>>;{ctrl+end}")
        time.sleep(4)
        actions.doAction('CLIPSAVE')
        t = unimacroutils.getClipboard()
        self.assert_equal("", t, "Clipboard should be empty now" ) 
        actions.doAction("<<copy>>")
        t = unimacroutils.getClipboard()
        self.assert_equal("", t, "Clipboard should still be empty" ) 
        testActionResult(0, "CLIPISNOTEMPTY")
        ## with empty clipboard restore goes automatically: 
        actions.doAction("CLIPRESTORE")
        t = unimacroutils.getClipboard()
        self.assert_equal("testing", t, "Clipboard should filled again" ) 

    def tttest_Non_Empty_clipboard_and_restore(self):
        testWindowContents = self.doTestWindowContents
        testActionResult = self.doTestActionResult
##         This test saves the clipboard, copies two letters
##         so the test CLIPISNOTEMPTY returns true
##         and the clipboard should be restored next
        actions.doKeystroke("testing")
        actions.doAction("<<selectall>><<copy>>{ctrl+end}")
        actions.doAction('CLIPSAVE')
        t = unimacroutils.getClipboard()
        self.assert_equal("", t, "Clipboard should be empty now" ) 
        actions.doAction("{shift+left 2}<<copy>>{ctrl+end}")
        t = unimacroutils.getClipboard()
        self.assert_equal("ng", t, "Clipboard should contain two letters now" ) 
        testActionResult(1, "CLIPISNOTEMPTY")
        actions.doAction("CLIPRESTORE")
        t = unimacroutils.getClipboard()
        self.assert_equal("testing", t, "Clipboard should filled now" ) 

    def tttest_complete_CLIP_action(self):
        testWindowContents = self.doTestWindowContents
        testActionResult = self.doTestActionResult
##         This test saves the clipboard, copies two letters
##         so the test CLIPISNOTEMPTY returns true
##         and the clipboard should be restored next
##         do not forget CLIPRESTORE!   
        actions.doKeystroke("testing")
        actions.doAction("<<selectall>><<copy>>{ctrl+end}")
        actions.doAction('CLIPSAVE; {shift+left 4}<<copy>>{ctrl+end}; CLIPISNOTEMPTY; {ctrl+end}abcd<<paste>>defg; CLIPRESTORE; <<paste>>')
        testWindowContents("testingabcdtingdefgtesting")

    def tttest_NON_complete_CLIP_action(self):
        testWindowContents = self.doTestWindowContents
        testActionResult = self.doTestActionResult
##         This test saves the clipboard,breaks off so does not return the ending
        actions.doKeystroke("testing")
        actions.doAction("<<selectall>><<copy>>{ctrl+end}")
        actions.doAction('CLIPSAVE; <<copy>>{ctrl+end}; CLIPISNOTEMPTY; {ctrl+end}abcd<<paste>>defg; CLIPRESTORE; <<paste>>')
        testWindowContents("testing")
        t = unimacroutils.getClipboard()
        # ??? x after testing in test procedure artefact::
        self.assert_equal("testing", t, "Clipboard should filled again now" ) 
        

# no main statement, run from command in _unimacrotest.py.




"""Read file with names (address book) and make nice list
for import in naturally speaking

testfile in unittest\namelist, some doctests in this file

possibilities for spoken forms, but not extensively elaborated.
the spoken forms for "-" seem to be not needed anymore.

Can be run from inputoutput.py (function namelist) or unimacro (function namelistUnimacro) 

"""
import re
from pathlib import Path
import string

# complete words that need a spoken form:
needsSpokenForm = {}
needsSpokenForm['kcs'] = 'kasee es'
needsSpokenForm['ksc'] = 'ka essee'
needsSpokenForm['ict'] = 'isee thee'
needsSpokenForm['rhd'] = 'er hadee'
needsSpokenForm['mwn'] = 'emwee en'
needsSpokenForm['mwu'] = 'emwee u'
needsSpokenForm['kb'] = 'kabee'
needsSpokenForm['bv'] = 'beevee'
needsSpokenForm["'t"] = "ut"

# parts of words that need a spoken form (combine in future with regularFormsDetails:
##spokenFormDetails = {}
##spokenFormDetails[chr(246) + "h"] = 'eu' # o umlaut + h
##spokenFormDetails[chr(246)] = 'u' # o umlaut
##spokenFormDetails["eij"] = 'ei'

# regular expressions that need spoken form:
regularForm = {}
regularExpr = {}
regularForm[r'sche$'] = 'se'
regularForm[r'^cha'] = 'sja'

# words that should not be capitalized, unless in the beginning of the name:
inBetweenWords = ['de', 'van']

def namelist(inputfile, outputfile):
    """Check filenames from input string end write good results to output

    "-" and spaces ok,

    ignore words with "_" or starting with number

    """
    firstNames = set()
    secondNames = set()
    fullNames = set()
    voornamenList = ['jan']
    for line in open(inputfile):
        tt = cleanLine(line)
        if not tt:
            continue
        tt = tt.split()
        firstName = ""
        secondName = ""
        if len(tt) < 1:
            continue
        if len(tt) == 1:
            secondName = tt
        elif len(tt) == 2:
            firstName = tt[:1]
            secondName = tt[1:]
        elif tt[1] in voornamenList:
            # len(tt) > 2
            firstName = tt[:2]
            secondName = tt[2:]
            #print 'voor  first:', firstName
            #print 'voor second:', secondName
        else:
            firstName = tt[:1]
            secondName = tt[1:]
            #print 'achter first:', firstName
            #print 'achtersecond:', secondName
        if firstName:
            firstName = WrittenSpoken(' '.join(firstName))
            firstNames.add(str(firstName))
        if secondName:
            secondName = WrittenSpoken(' '.join(secondName)) 
            secondNames.add(str(secondName))
        if firstName and secondName:
            fullName = firstName + secondName
            fullNames.add(str(fullName))
    total = [_f for _f in list(fullNames | firstNames | secondNames) if _f]
    total.sort()
    open(outputfile, 'w').write('\n'.join(total))

def namelistUnimacro(inputstring, ini=None):
    """Take a string as input, and get the several words lists as language dependent.

    Output a list of words.

    "-" and spaces ok,

    ignore words with "_" or starting with number

    """
    voornamenList = []
    inbetweenwords = []
    spokenForms = {}
    regularForms = {}
    result = []
    if ini:
        voornamenList =ini.getList("name phrase", "secondchristiannames") or []
        inbetweenwords = ini.getList("name phrase", "inbetweenwords") or []
        # spokenFormKeys = ini.get("spoken forms")
        for k in spokenForms:
            spokenForms[k] = ini.get("spoken forms", k)
##        regularFormKeys = ini.get("regular forms")
##        for k in regularFormKeys:
##            regularForms[k] = ini.get("regular forms", k)
    
    tt = cleanLine(inputstring)
    if not tt:
        return None
    tt = tt.split()
    firstName = ""
    secondName = ""
    if len(tt) < 1:
        return None
    if len(tt) == 1:
        secondName = tt
    elif len(tt) == 2:
        firstName = tt[:1]
        secondName = tt[1:]
    elif tt[1] in voornamenList:
        # len(tt) > 2
        firstName = tt[:2]
        secondName = tt[2:]
        #print 'voor  first:', firstName
        #print 'voor second:', secondName
    else:
        firstName = tt[:1]
        secondName = tt[1:]
        #print 'achter first:', firstName
        #print 'achtersecond:', secondName
    if firstName:
        firstName = WrittenSpoken(' '.join(firstName),
                                  spokenForms=spokenForms, regularForms=regularForms,
                                  inbetweenWords=inbetweenwords)
        result.append(str(firstName))
        
    if secondName:
        secondName = WrittenSpoken(' '.join(secondName),
                                   spokenForms=spokenForms, regularForms=regularForms,
                                   inbetweenWords=inbetweenwords) 
        result.append(str(secondName))
    if firstName and secondName:
        fullName = firstName + secondName
        result.append(str(fullName))
    return result


class WrittenSpoken:
    r"""Analyse the written\spoken for input

!!Note the doubling of backslashes, due to doctest functioning!    

Called with function above the needsSpokenForm and regularForm dicts as in this
file (for Dutch) are taken. From unimacro provide two dicts:
1. needsSpokenForms for words (abbrevs) that are translated completely
2. regularForms for patterns that are substituted (longest first)

FirstNames:
    
>>> str(WrittenSpoken("Dick"))
'Dick'
>>> str(WrittenSpoken("jan kees"))
'Jan Kees'
>>> str(WrittenSpoken("Jan-kees"))
'Jan-Kees'
>>> str(WrittenSpoken("Charaf-Marie"))
'Charaf-Marie\\\\sjaraf marie'

secondNames:
>>> str(WrittenSpoken("Jansen"))
'Jansen'
>>> str(WrittenSpoken("van Dijk-de Pater"))
'Van Dijk-de Pater'
>>> str(WrittenSpoken("van Dijk-De Graaf"))
'Van Dijk-De Graaf'

combinations:

>>> str(WrittenSpoken("Charaf de Graaf"))
'Charaf de Graaf\\\\sjaraf de graaf'

>>> str((WrittenSpoken("Charaf") + WrittenSpoken("de Leidsche")))
'Charaf de Leidsche\\\\sjaraf de leidse'

>>> str((WrittenSpoken("Kees") + WrittenSpoken("de Leidscher")))
'Kees de Leidscher'


    
    """

    def __init__(self, Input, spokenForms=None, regularForms=None, inbetweenWords=None):
        self.needsSpokenForm = spokenForms or needsSpokenForm
        self.regularForm = regularForms or regularForm
        self.inbetweenWords = inbetweenWords or inBetweenWords  # from formal parms or global (Dutch)
##        print 'inbetweenWords: %s'% self.inbetweenWords
##        print 'needsSpokenForm: %s'% self.needsSpokenForm
##        print 'regularForm: %s'% self.regularForm
        
        if isinstance(Input, WrittenSpoken):
            Input = str(Input)
        if Input.find('\\') > 0:
            self.written =  Input.split('\\')[0]
            self.spoken =  Input.split('\\')[ - 1]
        elif Input.find('-') > 0:
            parts = Input.split('-')
            results = list(map(WrittenSpoken, parts))
            written = [r.written for r in results]
            spoken = [r.spoken for r in results]
            self.written = '-'.join(written)
            self.spoken = ' '.join(spoken)
        elif Input.find(' ') > 0:
            parts = Input.split(' ')
            results = list(map(WrittenSpoken, parts))
            written = [r.written for r in results]
            spoken = [r.spoken for r in results]
            self.written = ' '.join(written)
            self.spoken = ' '.join(spoken)
        else:
            # capitalize NOT if in inbetweenWords:
            if Input in self.inbetweenWords:
                self.written = Input
            else:
                self.written = Input.capitalize()
            self.spoken = self.getSpokenForm(Input)
            
            if Input.lower() in needsSpokenForm:
                self.spoken = needsSpokenForm[Input.lower()]

    def __add__(self, other):
        new = WrittenSpoken(self)
        if not isinstance(other, WrittenSpoken):
            other = WrittenSpoken(other)
        new.written += ' ' + other.written
        new.spoken += ' ' + other.spoken
        return new
    

    def __str__(self):
        """note "-" is not regarded as reason for spoken form anymore!

        """
        writtenClean = self.written.lower().replace('-', ' ')
        spokenClean = self.spoken.lower()
        if writtenClean == spokenClean:
            result = self.written
        else:
            result = self.written  + '\\' + self.spoken
        if result[0] in string.ascii_lowercase:
            result = result[0].capitalize() + result[1:]
        return result

    def getSpokenForm(self, word):
        """should be a single word"""
        word = word.lower()
        # first the list of needsSpokenForm:
        if word in self.needsSpokenForm:
            return self.needsSpokenForm[word]
        decorated = [( - len(k), k) for k in self.regularForm]
        decorated.sort()
        regularFormKeys = [k for (dummy, k) in decorated]
        for k in regularFormKeys:
            if re.search(k, word):
                word = re.sub(k, self.regularForm[k], word)
                
        return word
    
              
def cleanLine(line):
    """Cleanout unwanted things, like " - " (must be "-")

    when line contains _ or starts with number return None, meaning a false entry

    """
    if line.find('-'):
        line = line.replace("- ", "-")
        line = line.replace(" -", "-")

    if line.find("_") >= 0:
        return None
    if line[0] in string.digits:
        return None
    return line

def _test():
    #pylint:disable=C0415
    import doctest
    doctest.master = None
    return  doctest.testmod()

def _testrun():
    # need files on computer QH for this, sorry...
    _dir = Path('C:/projects/unittest/namelist')
    infile = _dir/'names input.txt'
    outfile = _dir/'names output.txt'
    namelist(infile, outfile)

if __name__ == "__main__":
    _test()
    _testrun()



"""check the current state of the unimacro grammar files with the released versions

"""
#pylint:disable=R0912, R0914
import os
import shutil
import difflib
import subprocess
from pathlib import Path        
from natlinkcore import natlinkstatus

k3diffapp = r'C:\Program Files\KDiff3\kdiff3.exe'
if not os.path.isfile(k3diffapp):
    k3diffapp = None

status = natlinkstatus.NatlinkStatus()

    

def checkOriginalFileWithActualTxtPy(name, org_path, txt_path, py_path):
    """check if grammar has been copied, and changed, with copy of .txt as intermediate
    
    org_path: path to python file in UnimacroGrammars, the original grammars
    txt_path: initially copy of org_path, user area, ActiveGrammars, handled if new release has changes
    py_path:  actual state of active grammar, noted if changes are made
    
    """
    isfile = os.path.isfile
    if not isfile(txt_path):
        shutil.copyfile(org_path, txt_path)
    org_txt_equal = not bool( get_diff(org_path, txt_path) )
    
    if not isfile(py_path):
        if not org_txt_equal:
            print(f'\tnew release of not activated grammar {name}\n\t\tcopy {org_path} to\n\t\t\t{txt_path}')
            shutil.copyfile(org_path, txt_path)
        return 
    txt_py_equal = not bool( get_diff(txt_path, py_path) )
    if txt_py_equal:
        if org_txt_equal:
            # all equal
            return
        print(f'\tnew release of grammar file, copy to ActiveGrammars {name}')
        shutil.copyfile(org_path, txt_path)
        shutil.copyfile(txt_path, py_path)
        return
    # txt_py not equal
    if org_txt_equal:
        if have_symlinks:
            org_path_resolved = str(Path(org_path).resolve())
            print(f'****grammar "{name}" in (the active) UnimacroGrammars directory has been changed.')
            print(f'--------This new version "{py_path}" is copied to \n\t"{org_path_resolved}" and to\n\t"{txt_path}".')
            print(f'\tThe changes will be saved when you commit and push your git clone of unimacro, "{workDir}".')
            print(f'\t---If you want to undo your changes, revert in git ("{workDir}"),\n\tand copy manually back "{org_path_resolved}"\n\tto "{py_path}".\n--------')
            shutil.copyfile(py_path, org_path_resolved)
            shutil.copyfile(org_path_resolved, txt_path)
        else:
            print(f'****grammar {name} in ActiveGrammars changed, cannot copy to UnimacroGrammars because you are not developing in symlink mode (with "flit install --symlink")')
        return
    # changes AND new release:
    # one more possibility: org_path and py_path are identical, but txt_path is different:
    org_py_equal = not bool( get_diff(org_path, py_path) )
    if org_py_equal:
        shutil.copyfile(org_path, txt_path)
        return
    
    if have_symlinks:
        print(f'****changes of grammar {name}, both in UnimacroGrammars and ActiveGrammars\n\tcheck (yours): {py_path} and (release) {org_path}')
    else:
        print(f'****changes of grammar {name}, both in UnimacroGrammars and ActiveGrammars,\n\tbut beware, you are not in "symlink" mode (with "flit install --symlink")\n\t\tcheck (yours): {py_path} and (release) {org_path}')
    output_dir = Path(py_path).parent
    output_stem = "====" + Path(py_path).stem + "_diff"
    output_name = output_stem + ".txt"
    if k3diffapp:
        print_diff(org_path, py_path)
    else:
        diff_file = str(Path(output_dir)/output_name)
        print_diff(org_path, py_path, output_file=diff_file)


def get_diff(org_path, new_path):
    """print the diff of org and new
    """
    with open(org_path, 'r', encoding='utf-8') as org:
        with open(new_path, 'r', encoding='utf-8') as new:
            diff = difflib.unified_diff(
                org.readlines(),
                new.readlines(),
                fromfile=f'org: {org_path}',
                tofile=f'new: {new_path}',
            )
            return list(diff) or False

def print_diff(org_path, new_path, output_file=None):
    """do a diff and handle the result"""
    diff = get_diff(org_path, new_path)
    if not diff:
        return
            
    if output_file is None:
        if k3diffapp:
            subprocess.call([k3diffapp, org_path, new_path])
            return
        for line in diff:
            print(line)
        return
    with open(output_file, 'w', encoding='utf-8') as fp:
        fp.writelines(diff)
    print(f'----diff in {output_file}')

def cleanup_files(activedir):
    """remove diff files starting with "===="
    """
    cleanupFiles = [f for f in os.listdir(activedir) if f.startswith('====')]
    for f in cleanupFiles:
        os.remove(os.path.join(activedir, f))

def checkUnimacroGrammars():
    """see if there are any changed grammar files with respect to original file in release
    
    sync with ...
    """
    join, isdir, isfile, listdir = os.path.join, os.path.isdir, os.path.isfile, os.listdir
    u_dir = status.getUnimacroDirectory()
    # u_user_dir = status.getUnimacroUserDirectory()
    u_grammars_dir = status.getUnimacroGrammarsDirectory()
    u_original_grammars_dir = join(u_dir, "UnimacroGrammars")
    assert isdir(u_original_grammars_dir)
    originalPyFiles = [f for f in listdir(u_original_grammars_dir) if f.endswith('.py')]
    txtFiles = [f for f in listdir(u_grammars_dir) if f.endswith('.txt')]
    activePyFiles = [f for f in listdir(u_grammars_dir) if f.endswith('.py')]
    
    for f in originalPyFiles:
        org_path = join(u_original_grammars_dir, f)
        txt_file = f.replace('.py', '.txt')
        txt_path = join(u_grammars_dir, txt_file)
        py_path = join(u_grammars_dir, f)
        nice_name = f[:-3]   # strip off .py
        checkOriginalFileWithActualTxtPy(nice_name, org_path, txt_path, py_path)
        
    for f in txtFiles:
        f_py = f.replace('.txt', '.py')
        if f_py not in originalPyFiles:
            print(f'txt file "{f}" in ActiveGrammars, but py file {f_py} not in UnimacroGrammars')
            os.remove(f)
            
    for f in activePyFiles:
        f_txt = f.replace('.py', '.txt')
        if not isfile(f_txt):
            if f in originalPyFiles:
                nice_name = f[:-3]   # strip off .py
                org_path = join(u_original_grammars_dir, f)
                txt_path = join(u_grammars_dir, f_txt)
                py_path = join(u_grammars_dir, f)
                checkOriginalFileWithActualTxtPy(nice_name, org_path, txt_path, py_path)

        
if __name__ == "__main__":
    checkUnimacroGrammars()   




#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
from dtactions.unimacro import unimacroutils
from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroactions as actions
actions.debugActions(1)

import unittest
import UnimacroTestHelpers
import time
import sys

##class UnimacroBasicTest(TestCaseWithHelpers.TestCaseWithHelpers):
class SearchTest(UnimacroTestHelpers.UnimacroTestHelpers):
    """Tests several search actions.

    executes on the active window which should be empty before starting!   

    """
      
    def setUp(self):
        self.doTestWindowIsEmpty(testName="setUp SearchTest needs an empty window to start with...")

    def tearDown(self):
        pass

    def getWindowContents(self):
        action = actions.doAction
        action('CLIPSAVE')
        action("<<selectall>><<copy>>")
        contents = unimacroutils.getClipboard()
        action('CLIPRESTORE')
        return contents

    def doTestClipboard(self, expected, testName=None):
        text = 'testing the clipboard'
        if testName:
            text += ': ' + testName
        action = actions.doAction
        contents = unimacroutils.getClipboard()
        self.assert_equal(expected, contents, text)


    def doTestActionResult(self, expected, commandString):
        result = Search.doAction(commandString)
        if expected and result:
            pass
        elif not expected and not result:
            pass
        else:
            self.assert_equal(expected, result, "result of action not as expected")



    def checkContents(self, expected, comment=""):
        """check expected tuple (clipboard, before, after)

        if selection was active, it came through a forward search, feeble part!
        
        """
        text = 'checking contents'
        if comment:
            text += ': %s'% comment
        action = actions.doAction
        action('CLIPSAVE')
        action("<<copy>>")
        clip = unimacroutils.getClipboard()
        lenClip = len(clip)
        # get to the left of the selection:
        action('<<leftafterforwardsearch %s>>{shift+ctrl+home}<<copy>>{ctrl+home}'% lenClip)
        begin = unimacroutils.getClipboard()
        lenBegin = len(begin)
        lengths = lenBegin + lenClip
        action('{right %s}'% lengths)
        action('{shift+ctrl+end}<<copy>>')
        end = unimacroutils.getClipboard()
        end = end.rstrip()
        lenEnd = len(end)
        action('<<selectall>><<copy>>')
        all = unimacroutils.getClipboard()
        all = all.rstrip()
        lenAll = len(all)
        action('CLIPRESTORE')
        self.assertEqual(lenAll, lenBegin+lenClip+lenEnd,
                          text + 'lengths do not match, total: %s, parts: %s, %s, %s'%
                          (lenAll, lenBegin, lenClip, lenEnd))
        action('VW')
        action("{ctrl+home}{right %s}{shift+right %s}"% (lenBegin, lenClip))
        self.assert_equal( expected, (begin, clip, end), text)    

    

        
    def test_basic_Search_actions(self):
        action = actions.doAction
        keystroke = actions.doKeystroke
        keystroke('aaa\nbb\n   ccc')
        action("{ctrl+home}")
        expected = ("", "", 'aaa\nbb\n   ccc')  # selection, before, after
        # if this test fails, one of the actions of "checkContents" does not work properly:
        self.checkContents(expected, comment="initial position")

        keystroke("{right}{shift+right}")
        expected = ("a", "a", 'a\nbb\n   ccc')  # selection, before, after
        self.checkContents(expected, comment="middle a selected")
        
        
        action("<<startsearch>>")
        keystroke("bb")
        action("<<searchgo>>")
        action("<<copy>>")
        self.doTestClipboard("bb", "clipboard contents not as expected")
        expected = ('aaa\n', 'bb', '\n   ccc')  # selection, before, after
        self.checkContents(expected, comment="after searchfor bb action ")        
        


# no main statement, run from command in _unimacrotest.py.




# (unimacro - natlink macro wrapper/extensions)
# (c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)
#                    Ben Staniford (ben_staniford@users.sourceforge.net)
#                    Bart Jan van Os (bjvo@users.sourceforge.net)
#
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net).
#
# "unimacro" is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, see:
# http://www.gnu.org/licenses/gpl.txt
#
# "unimacro" is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; See the GNU General Public License details.
#
# "unimacro" makes use of another SourceForge project "natlink",
# which has the following copyright notice:
#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# June 2020: adapt for python3, Quintijn Hoogenboom
# Author: Bart Jan van Os, Version: 1.0
#pylint:disable = C0209, W0622, C0321, W0612, W0702
#pylint:disable = E1101
"""This file implements a dialog/window to browse and train NatLink
grammars

The browser is based upon a tree dialog, adapted from the hiertest demo

Note: You can only select and train items in the right pane to avoid
accidently training large collections of words

Simply start the python win gui and run this module, to see an example.
An explanation of how to use this browser with your own grammars can be
found in _grammarUtil.py, BrowseGrammar.py and natlinkutilsbj.py The
_grammarUtil module has a method for building the required grammar.bin
files.

This file is called when doing the command "Show All Grammars" from the
grammar "_control".

"""
import sys
import pickle
import win32ui
import win32con
import win32api
import commctrl
from pywin.mfc import dialog
from pywin.tools import hierlist
from pywin.framework import dlgappcore

import natlink
from natlinkcore.natlinkutils import *
from natlinkcore import natlinkstatus
from natlinkutilsbj import SetMic,GrammarFileName

from BrowseGrammar import *
from listdialogs import ListDialog
import TrainDialog
import D_train

status = natlinkstatus.NatlinkStatus()
dataDirectory = status.getUnimacroDataDirectory()

# seems to have gone from comctrl:
LVCFMT_LEFT = 0


class GramHierList(hierlist.HierList):
    def __init__(self, root,Start, listBoxID = win32ui.IDC_LIST1):
        hierlist.HierList.__init__(self, root, win32ui.IDB_BROWSER_HIER, listBoxID)
        self.Start=Start

    def GetText(self, item):
        return item.GetName()
    
    def GetSubList(self, item):
        if item.GramType==RuleCode:
            InnerRules=item.GetInnerRules(0)
        else:
            InnerRules = None
        return InnerRules
    
    def IsExpandable(self, item):
        return (item.GramType==RuleCode) and not item.AreAllWords()
    
    def GetSelectedBitmapColumn(self, item):
        return self.GetBitmapColumn(item)+6 # Use different color for selection

    def TakeDefaultAction(self, item):
        # Restore original behavior :-)
        # better is to overwrite and restore original behavior of OnTreeItemDoubleClick
        if self.IsExpandable(item):
            self.Expand(self.GetSelectedItem(),commctrl.TVE_TOGGLE)

    def OpenNext(self,root,i,RulePath):
        if i<len(RulePath):
            for o in root.Included:
                if not IsText(o):
                    if id(o)==id(RulePath[i]):
                        if self.IsExpandable(o):
                            self.Expand(self.GetHandleOf(o),commctrl.TVE_EXPAND)                        
                        if i==len(RulePath)-1:
                            self.SelectItem(self.GetHandleOf(o))                            
                        else:
                            self.OpenNext(o,i+1,RulePath)

    def OpenRule(self,RulePath):
        self.OpenNext(self.root,0,RulePath)

    def OpenStart(self):
        if len(self.Start) == 2:
            _Rule, _Path, objPath=self.root.FindRulePath(self.Start)
            for o in objPath:
                if not IsText(o):
                    self.OpenRule(objPath)

    def GetHandleOf(self,item):
        HandlesFor=InverseDict(self.itemHandleMap)
        if item in HandlesFor:
            return HandlesFor[item]
        return None


IDC_EDIT=1000
IDC_SYNTAX=1010
IDC_TRAIN=1011
IDC_TRAINSPECIAL=1012
IDC_EDITITEM=1013
IDC_NEWITEM=1014

class GrammarDialog(dialog.Dialog):
#    def __init__(self, title, hierList, bitmapID = win32ui.IDB_HIERFOLDERS, dlgID = win32ui.IDD_TREE, dll = None, childListBoxID = win32ui.IDC_LIST1):
    def __init__(self, title, hierList):
        dialog.Dialog.__init__ (self, self._maketemplate(title))
        self.hierList=hierList
        self.colHeadings=['Name','Definition']
        self.items=[]
        self.LastFocused=0
        self.SyntaxItem=0

    def _maketemplate(self,title):
        style = (win32con.WS_DLGFRAME | win32con.WS_MAXIMIZEBOX |
                win32con.WS_MINIMIZEBOX | win32con.WS_SIZEBOX |
                win32con.WS_SYSMENU | win32con.WS_VISIBLE |
                 win32con.WS_EX_TOPMOST 
                 )
        cs = (win32con.WS_CHILD           | win32con.WS_TABSTOP |
            win32con.WS_VISIBLE         | commctrl.CCS_NOMOVEY |
            commctrl.TVS_HASLINES       | commctrl.TVS_SHOWSELALWAYS |
            commctrl.TVS_LINESATROOT    | win32con.WS_BORDER |
            commctrl.TVS_HASBUTTONS)
        ls = (
            win32con.WS_CHILD           | win32con.WS_BORDER |
            win32con.WS_VISIBLE         | win32con.WS_TABSTOP |
            commctrl.LVS_ALIGNLEFT      |
            commctrl.LVS_REPORT         | commctrl.LVS_EDITLABELS
            )
        t = [ [title, (0, 0, 500, 300), style, None, (8, "MS Sans Serif")],
            ["SysTreeView32", None, win32ui.IDC_LIST1, (0, 0, 1, 1), cs],
            ["SysListView32", None, IDC_SYNTAX, (0, 0, 1, 1), ls],
            ]
        return t

    #(only for DlgAppCore)
    def PreDoModal(self):
        sys.stdout = sys.stderr = self


    def OnInitDialog(self):
        self.CenterWindow()                    
        self.hierList.HierInit(self)
        rc=dialog.Dialog.OnInitDialog(self)
        self.HookMessage (self.on_size, win32con.WM_SIZE)
        self.HookNotify(self.SyntaxItemChange, commctrl.LVN_ITEMCHANGED)
        self.HookNotify(self.TreeItemChange, commctrl.TVN_SELCHANGED)
        self.HookNotify(self.GotFocus, commctrl.NM_SETFOCUS)
        self.HookCommand(self.OnSyntaxClick, IDC_SYNTAX)
        self.Tree = self.GetDlgItem(win32ui.IDC_LIST1)
        self.Syntax = self.GetDlgItem(IDC_SYNTAX)
        self.FillList()
        size = self.GetWindowRect()
        self.LayoutControls(size[2]-size[0], size[3]-size[1])
        self.hierList.OpenStart()
        self.SetFocus()
        return rc

    def ExpandSyntaxItem(self,ItemNr):
        Item=self.items[ItemNr]
        Name=Item[0]
        State=self.Syntax.GetItemState(ItemNr,commctrl.LVIS_SELECTED)        
        if (Name!='<SpokenForm>') and (Name!='<Syntax>') and (State!=0):
            # force parent Tree item to expand, adding the sublist to the itemHandle map
            hItem=self.hierList.GetSelectedItem()
            oItem=self.hierList.ItemFromHandle(hItem)            
            if self.hierList.IsExpandable(oItem):
                self.hierList.Expand(hItem,commctrl.TVM_EXPAND)
            # If it exists, select the Tree item corresponding to the Syntax item
            h=self.hierList.GetHandleOf(Item[2])
            if h: self.Tree.SelectItem(h)

    def OnSyntaxClick(self, id, code):
        if code==commctrl.NM_DBLCLK:
            self.ExpandSyntaxItem(self.SyntaxItem)
        return 1

    def OnOK(self):
        if self.LastFocused==IDC_SYNTAX:
            self.ExpandSyntaxItem(self.SyntaxItem)

    def GotFocus(self,std, extra):
        (hwndFrom, idFrom, code)= std
        self.LastFocused=idFrom

    def on_size (self, params):
        lparam = params[3]
        w = win32api.LOWORD(lparam)
        h = win32api.HIWORD(lparam)
        self.LayoutControls(w, h)
    
    def LayoutControls(self, w, h):
        d=int(w/4)
        self.Tree.MoveWindow((0,0,d,h))
        self.Syntax.MoveWindow((d,0,w,h))

    def SyntaxItemChange(self,std, extra):
        (hwndFrom, idFrom, code), (itemNotify, sub, newState, oldState, change, point, lparam) = std, extra
        oldSel = (oldState & commctrl.LVIS_SELECTED)!=0
        newSel = (newState & commctrl.LVIS_SELECTED)!=0
        if oldSel != newSel:
            self.SyntaxItem=itemNotify


    def SetRuleItems(self,Rule,InnerRules):
        IsGrammar=Rule.IsRuleContainer()
        FoldedContents=Rule.GetContents(1,Unfold=0)
        IsSimple=Rule.AreAllWords(FoldedContents)
        Items=[]
        if not (IsGrammar or IsSimple): Items.append(('<Syntax>',FoldedContents,Rule))
        for IR in InnerRules:
            if IR.IsAltOrList():
                Items.append((IR.GetName(),IR.GetContents(1,Unfold=1),IR))
            else:
                Items.append((IR.GetName(),IR.GetContents(1,Unfold=0),IR))
        if not IsGrammar:
            Items.append(('<SpokenForm>',Rule.GetContents(2,Unfold=1),Rule))
        self.items=Items

    def SetAlternativesItems(self,Rule,InnerRules):
        Items=[]
        Alternatives=Rule.Included
        Alternatives.sort(key=str.lower)
        IgnoreBracketsParen=re.compile('\[|\]|\(|\)')        
        for a in Alternatives:
            s=a.strip()
            s=IgnoreBracketsParen.sub('',s)
            if s in Rule.AlternativesDict:
                Items.append((s,Rule.AlternativesDict[s],Rule,1))
            else:
                Items.append((s,'',Rule,1))
        self.items=Items


    def SetListItems(self,Rule):
        Items=[]
        Elements=Rule.Included
        Elements.sort(key=str.lower)
        for a in Elements:
            if a in Rule.AlternativesDict:
                Items.append((a,Rule.AlternativesDict[a],Rule,1))
            else:
                Items.append((a,'',Rule,1))
        if Elements==[]:
            Items.append(('{Unknown list items}','',Rule))
        self.items=Items
        
        
    def TreeItemChange(self,std, extra):
        hwndFrom, idFrom, code=std
        Action,ItemOld,ItemNew,Point=extra
        Rule=self.hierList.ItemFromHandle(self.Tree.GetSelectedItem())
        InnerRules=Rule.GetInnerRules(0)
        if Rule.GramType==RuleCode:
            self.SetRuleItems(Rule,InnerRules)
        elif Rule.IsLongAlternative():
            self.SetAlternativesItems(Rule,InnerRules)
        elif Rule.GramType==ListCode:
            self.SetListItems(Rule)
        else:
            pass
        self.InsertItems()
       

    def InsertItems(self):
        self.Syntax.DeleteAllItems()
        index=0
        for items in self.items:
            index = self.Syntax.InsertItem(index+1, str(items[0]), 0)
            for itemno in range(1,len(self.colHeadings)):
                item = items[itemno]
                self.Syntax.SetItemText(index, itemno, str(item))
        try:
            self.Syntax.SetItemState(0, commctrl.LVIS_FOCUSED+commctrl.LVNI_SELECTED, 255)
        except:
            pass
        
    def FillList(self):
        size = self.GetWindowRect()
        width = 2*(size[2] - size[0])/3 - win32api.GetSystemMetrics(win32con.SM_CXVSCROLL)
        colw=[0.35,2.80]
        numCols = len(self.colHeadings)
        index = 0
        for col in self.colHeadings:
            ## change second variable to int, QH, 23092022::
            itemDetails = (LVCFMT_LEFT, int(width*colw[index]), col, 0)
            print(f'index: {index}, itemDetails: {itemDetails}')
            self.Syntax.InsertColumn(index, itemDetails)
            index = index + 1
        index = 0


class TrainGrammarDialog(GrammarDialog):
#    def __init__(self, title, hierList, bitmapID = win32ui.IDB_HIERFOLDERS, dlgID = win32ui.IDD_TREE, dll = None, childListBoxID = win32ui.IDC_LIST1):
    def __init__(self, title, hierList):
        dialog.Dialog.__init__ (self, self._maketemplate(title))
        self.hierList=hierList
        self.colHeadings=['Name','Definition']
        self.items=[]
        self.LastFocused=0
        self.SelItems=[]
        

    def _maketemplate(self,title):
        style2 = win32con.WS_OVERLAPPEDWINDOW | win32con.WS_VISIBLE        
        style = (win32con.WS_DLGFRAME | win32con.WS_MAXIMIZEBOX |
                win32con.WS_MINIMIZEBOX | win32con.WS_SIZEBOX |
                win32con.WS_SYSMENU | win32con.WS_VISIBLE |
                 win32con.SW_SHOW | win32con.WS_EX_TOPMOST 
                 )
        child = win32con.WS_CHILD | win32con.WS_VISIBLE
        cs = (child           |
            win32con.WS_TABSTOP         | commctrl.CCS_NOMOVEY |
            commctrl.TVS_HASLINES       | commctrl.TVS_SHOWSELALWAYS |
            commctrl.TVS_LINESATROOT    | win32con.WS_BORDER |
            commctrl.TVS_HASBUTTONS)
        ls = (
            win32con.WS_CHILD           | win32con.WS_BORDER |
            win32con.WS_VISIBLE         | win32con.WS_TABSTOP |
            commctrl.LVS_ALIGNLEFT      |
            commctrl.LVS_REPORT         | commctrl.LVS_EDITLABELS
            )
        bs = (
            win32con.WS_CHILD           |
            win32con.WS_VISIBLE
             )
        t = [ [title, (0, 0, 500, 300), style, None, (8, "MS Sans Serif")],
            ["SysTreeView32", None, win32ui.IDC_LIST1, (0, 0, 1, 1), cs],
            ["SysListView32", None, IDC_SYNTAX, (0, 0, 1, 1), ls],
            [128, "Exit", win32con.IDCANCEL, (10, 0, 50, 14), bs ],
            [128, "Train Special", IDC_TRAINSPECIAL, (10, 0, 50, 14), bs ],                                          
            [128, "&Train", IDC_TRAIN, (10, 0, 50, 14), bs | win32con.BS_PUSHBUTTON],
            [128, "&Edit", IDC_EDITITEM, (10, 0, 50, 14), bs ],              
            [128, "&New", IDC_NEWITEM, (10, 0, 50, 14), bs ],              
            [129,"",IDC_EDIT,(7,72,250,85),child|win32con.ES_MULTILINE|win32con.WS_BORDER|
                    win32con.ES_AUTOVSCROLL|win32con.WS_VSCROLL],
            ]
        return t

    def OnInitDialog(self):
        self.CenterWindow()                    
        self.hierList.HierInit(self)
        rc=dialog.Dialog.OnInitDialog(self)
        self.HookMessage (self.on_size, win32con.WM_SIZE)
        self.HookNotify(self.SyntaxItemChange, commctrl.LVN_ITEMCHANGED)
        self.HookNotify(self.TreeItemChange, commctrl.TVN_SELCHANGED)
        self.HookNotify(self.GotFocus, commctrl.NM_SETFOCUS)
        self.HookCommand(self.OnSyntaxClick, IDC_SYNTAX)
        self.HookCommand(self.onTrain, IDC_TRAIN)
        self.HookCommand(self.onTrainSpecial, IDC_TRAINSPECIAL)
        self.HookCommand(self.onNew, IDC_NEWITEM)
        self.HookCommand(self.onEdit, IDC_EDITITEM)
        self.Tree = self.GetDlgItem(win32ui.IDC_LIST1)
        self.Syntax = self.GetDlgItem(IDC_SYNTAX)
        self.butTrainSpecial = self.GetDlgItem(IDC_TRAINSPECIAL)
        self.butTrain = self.GetDlgItem(IDC_TRAIN)
        self.butCancel = self.GetDlgItem(win32con.IDCANCEL)
        self.butEdit = self.GetDlgItem(IDC_NEWITEM)
        self.butNew = self.GetDlgItem(IDC_EDITITEM)        
        self.Output= self.GetDlgItem(IDC_EDIT)
        self.oldStdout = sys.stdout        
        sys.stdout = self
        self.message = ''
        
        self.FillList()
        size = self.GetWindowRect()
        self.LayoutControls(size[2]-size[0], size[3]-size[1])
        self.hierList.OpenStart()
        self.SetFocus()
        rc=GrammarDialog.OnInitDialog(self)
        return rc

    def OnDestroy(self,msg):
        sys.stdout = self.oldStdout
    
    def write(self,text):
        self.message = self.message + text.replace('\n','\r\n')
        self.SetDlgItemText(IDC_EDIT,self.message)
        self.GetDlgItem(IDC_EDIT).SendMessage(win32con.EM_SETSEL,0x7FFF,0x7FFF)
        self.GetDlgItem(IDC_EDIT).SendMessage(win32con.EM_SCROLLCARET,0,0)

    def LayoutControls(self, w, h):
        d=w//4
        self.Tree.MoveWindow((0,0,d,h-100))
        self.Syntax.MoveWindow((d,0,w,h-100))
        self.Output.MoveWindow((0,h-97,w,h-30))
        self.butCancel.MoveWindow((w-60, h-24, w-10, h-4))
        self.butTrain.MoveWindow((w-120, h-24, w-70, h-4))
        self.butEdit.MoveWindow((w-180, h-24, w-130, h-4))
        self.butNew.MoveWindow((w-240, h-24, w-190, h-4))                
        self.butTrainSpecial.MoveWindow((10, h-24, 90, h-4))

    def GotFocus(self,std, extra):
        (hwndFrom, idFrom, code)= std
        self.LastFocused=idFrom
        if idFrom==IDC_SYNTAX:
            self.butTrain.EnableWindow(1)
            self.butEdit.EnableWindow(0)
            self.butNew.EnableWindow(0)            
        else:
            self.butTrain.EnableWindow(0)
            self.butEdit.EnableWindow(0)
            self.butNew.EnableWindow(0)            
            

    def SyntaxItemChange(self,std, extra):
        (hwndFrom, idFrom, code), (itemNotify, sub, newState, oldState, change, point, lparam) = std, extra
        oldSel = (oldState & commctrl.LVIS_SELECTED)!=0
        newSel = (newState & commctrl.LVIS_SELECTED)!=0
        if oldSel != newSel:
            try:
                if newSel:
                    self.SelItems.append(itemNotify)
                else:
                    if itemNotify in self.SelItems:
                        self.SelItems.remove(itemNotify)
            except win32ui.error:
                self.SelItems=[]

    def OnSyntaxClick(self, id, code):
        if code==commctrl.NM_DBLCLK:
            if len(self.SelItems)==1:
                self.ExpandSyntaxItem(self.SelItems[0])
        return 1

    def InsertItems(self):
        GrammarDialog.InsertItems(self)
        self.SelItems=[0]

    def OnOK(self):
        if self.LastFocused==IDC_SYNTAX:
            if len(self.SelItems)==1:
                self.ExpandSyntaxItem(self.SelItems[0])

    def GetTextChunks(self):
        InnerRules=[]
        ToTrain=[]
        Rule=self.items[0][2]
        for Nr in self.SelItems:
            if len(self.items[Nr])>3:
                ToTrain.append(self.items[Nr][0])
            else:
                Rule=self.items[Nr][2]
                InnerRules.append(Rule)
                InnerRules.extend(Rule.GetInnerRules(10))
        InnerRules=Rule.RemoveDuplicates(InnerRules)
        for Rule in InnerRules:
            ToTrain.extend(Rule.GetTextChunks())
        ToTrain.sort(key=str.lower)
        RemoveDuplicatesOfSortedList(ToTrain)
        if ToTrain[0]=='<imported>': ToTrain.remove('<imported>')
        return ToTrain
    

    def onTrain(self,nID,code):
        #natlink.recognitionMimic(['Print','Module','Info'])
        #return
        if (self.LastFocused==IDC_SYNTAX) and len(self.SelItems)>0:
            Name1=self.items[self.SelItems[0]][0]
            if (Name1!='<Syntax>'):
                ToTrain=self.GetTextChunks()
                self.goTrain(ToTrain)

    def onTrainSpecial(self,nID,code):
        #pylint:disable=W0101
        return  ## just disable...
        Names=[]
        Keys=list(D_train.SpecialTraining.keys())
        for name in Keys:
            Names.append((name))
        (x,y)=natlink.getScreenSize()
        Size=(int(x*0.20),int(y*0.30))
        dlg = ListDialog('Select Special Training', Names, 'Grammar',size=Size,okButton='Train')
        if dlg.DoModal()==win32con.IDOK:
            Selection=dlg.Selection
            ToTrain=[]
            for k in Selection:
                ToTrain.extend(D_train.SpecialTraining[k])
            ToTrain.sort(key=str.lower)
            RemoveDuplicatesOfSortedList(ToTrain)
            self.goTrain(ToTrain)

    def onNew(self,nID,code):
        pass

    def onEdit(self,nID,code):
        #pylint:disable=W0101
        return
        if self.LastFocused==IDC_SYNTAX:
            if len(self.SelItems)==1:
                print(self.SelItems[0])
                self.EditControl=self.Syntax.EditLabel(0)
                #self.Syntax.SetFocus()
                #self.Syntax.EditLabel(self.SelItems[0])        
    

    def goTrain(self,ToTrain):
        Title='Train Phrases'
        Phrases=[]
        for p in ToTrain:
            Phrases.append((p))
        try:
            natlink.natConnect(1)
        except:
            return
        MicState=natlink.getMicState()
        dlg=TrainDialog.TrainDialog(Title,Phrases,MicState)
        SetMic('off')
        dlg.DoModal()
        natlink.natDisconnect()    

                        
class BrowseDialogApp(dlgappcore.DialogApp):
    "BJ added: An application class, for an app with main dialog box"
    def InitInstance(self):
#       win32ui.SetProfileFileName('dlgapp.ini')
#        win32ui.LoadStdProfileSettings()
        win32ui.EnableControlContainer()
        win32ui.Enable3dControls()
        self.dlg = self.frame = self.CreateDialog()
        if self.frame is None:
            raise Exception("No dialog was created by CreateDialog()")
        self.PreDoModal()
        self.dlg.PreDoModal()
        self.dlg.DoModal()
        self.PostDoModal()
    
    def CreateDialog(self):
        return CreateBrowseDialog()

    def PreDoModal(self):
        pass
        #natlink.natConnect(1) #presently handled in TrainDialog

    def PostDoModal(self):
        pass
        #natlink.natDisconnect()

def CreateBrowseDialog():
    try:
        GrammarFile=open(GrammarFileName,'rb')
    except:
        GrammarFile=open(dataDirectory+'\\TestGrammar.bin','rb')
    (Grammars,Start,All,Exclusive)=pickle.load(GrammarFile)
    GrammarFile.close()
    Grammars.Sort()
    if Exclusive: Name='Exclusive Grammars (Active Rules)'
    elif All: Name='All Grammars'
    else:   Name='Active Grammars'
    dlg=GrammarDialog(Name,GramHierList(Grammars,Start))
    # dlg=TrainGrammarDialog(Name,GramHierList(Grammars,Start))
    return dlg


def demomodeless():
    dlg=CreateBrowseDialog()
    dlg.CreateWindow()

def demodlg ():
    dlg=CreateBrowseDialog()    
    dlg.DoModal()


def CheckCreateApp():
    if "pywin.framework.startup" in sys.modules:
        App=BrowseDialogApp()


if __name__=='__main__':
    ## assume there is a "grammar.bin" in the UnimacroDataDirectory 
    CheckCreateApp()
    # 
    # 
    # demodlg()
    # #demomodeless()
else:
    CheckCreateApp()
    